import numpy as np
import pytest
import sys

from packaging import version as pversion

import psyneulink.core.components.functions.nonstateful.distributionfunctions as Functions
from psyneulink.core.globals.utilities import _SeededPhilox

np.random.seed(0)
test_var = np.random.rand()

RAND1 = np.random.rand()
RAND2 = np.random.rand()
RAND3 = np.random.rand()
RAND4 = np.random.rand()
RAND5 = np.random.rand()

dda_expected_default = (1.9774974807292212, 0.012242689689501842,
                        1.9774974807292207, 1.3147677945132479, 1.7929299891370192,
                        1.9774974807292207, 1.3147677945132479, 1.7929299891370192)
dda_expected_random = (0.4236547993389047, -2.7755575615628914e-17,
                       0.5173675420165031, 0.06942854144616283, 6.302631815990666,
                       1.4934079600147951, 0.4288991185241868, 1.7740760781361433)
dda_expected_negative = (0.42365479933890504, 0.0,
                         0.5173675420165031, 0.06942854144616283, 6.302631815990666,
                         1.4934079600147951, 0.4288991185241868, 1.7740760781361433)
dda_expected_small = (0.5828813465336954, 0.04801236718458773,
                      0.532471083815943, 0.09633801555720854, 6.1142591416669765,
                      1.5821207676710864, 0.5392724051148722, 1.806647390875747)

# Different libm implementations produce slightly different results
# Numpy 1.22+ uses new/optimized implementation of FP routines
# on processors that support AVX512 since 1.22 [0]
# [0] https://github.com/numpy/numpy/commit/1eff1c543a8f1e9d7ea29182b8c76db5a2efc3c2
if sys.platform.startswith("win") or \
   sys.platform.startswith("darwin") or \
   (pversion.parse(np.version.version) >= pversion.parse('1.22') and pytest.helpers.numpy_uses_avx512()):
    dda_expected_small = (0.5828813465336954, 0.04801236718458773,
                          0.5324710838150166, 0.09633802135385469, 6.117763080882898,
                          1.58212076767016, 0.5392724012504414, 1.8064031532265)

# Numpy 1.23+ reimplemented tanh to roughly follow SVML but not enough to match
# the above results [1,2]
# [1] https://github.com/numpy/numpy/pull/20363
# [2] https://github.com/numpy/numpy/commit/75edab9f7a7d95ecc62efc1b3b92642b6d45762d
elif pversion.parse(np.version.version) >= pversion.parse('1.23'):
    dda_expected_small = (0.5828813465336954, 0.04801236718458773,
                          0.5324710838150166, 0.09633801169277778, 6.111024594252574,
                          1.58212076767016, 0.5392724012504414, 1.8064031532265)

normal_expected_mt = (1.0890232855122397)
uniform_expected_mt = (0.6879771504250405)
normal_expected_philox = (0.5910357654927911)
uniform_expected_philox = (0.6043448764869507)

llvm_expected = {'fp64': {}, 'fp32': {}}

# LLVM computes slightly different results from Python/numpy even for fp64
llvm_expected['fp64'][dda_expected_small] = (0.5828813465336954, 0.04801236718458773,
                                             0.5324710838085324, 0.09633788030213193, 6.0183026674990625,
                                             1.5821207675877176, 0.5392731084412705, 1.843608020219776)

# DDA fp32 results are noticeably different due to lower precision.
llvm_expected['fp32'][dda_expected_default] = (1.9774975776672363, 0.012242687866091728,
                                               1.9774975776672363, 1.3147677183151245, 1.792930245399475,
                                               1.9774975776672363, 1.3147677183151245, 1.792930245399475)
llvm_expected['fp32'][dda_expected_random] = (0.42365485429763794, 0.0,
                                              0.5173675417900085, 0.069428451359272, 6.302595138549805,
                                              1.4934077262878418, 0.42889538407325745, 1.7739042043685913)
llvm_expected['fp32'][dda_expected_negative] = (0.4236549735069275, 5.960464477539063e-08,
                                                0.5173678398132324, 0.06942932307720184, 6.302994251251221,
                                                1.4934080839157104, 0.4288962781429291, 1.7739406824111938)
llvm_expected['fp32'][dda_expected_small] = None

# Philox uses different algorithm for fp32 (consumes fewer bits of entropy)
llvm_expected['fp32'][normal_expected_philox] = (0.5655658841133118)
llvm_expected['fp32'][uniform_expected_philox] = (0.6180108785629272)

test_data = [
    pytest.param(Functions.DriftDiffusionAnalytical, test_var, {}, None, None,
                 dda_expected_default, id="DriftDiffusionAnalytical-DefaultParameters"),
    pytest.param(Functions.DriftDiffusionAnalytical, test_var,
                 {"drift_rate": RAND1, "threshold": RAND2, "starting_value": RAND3,
                  "non_decision_time":RAND4, "noise": RAND5}, None, None,
                 dda_expected_random, id="DriftDiffusionAnalytical-RandomParameters"),
    pytest.param(Functions.DriftDiffusionAnalytical, -test_var,
                 {"drift_rate": RAND1, "threshold": RAND2, "starting_value": RAND3,
                  "non_decision_time":RAND4, "noise": RAND5}, None, None,
                 dda_expected_negative, id="DriftDiffusionAnalytical-NegInput"),
    pytest.param(Functions.DriftDiffusionAnalytical, 1e-4,
                 {"drift_rate": 1e-5, "threshold": RAND2, "starting_value": RAND3,
                  "non_decision_time":RAND4, "noise": RAND5}, None, "Rounding Errors",
                 dda_expected_small, id="DriftDiffusionAnalytical-SmallDriftRate"),
    pytest.param(Functions.DriftDiffusionAnalytical, -1e-4,
                 {"drift_rate": 1e-5, "threshold": RAND2, "starting_value": RAND3,
                  "non_decision_time":RAND4, "noise": RAND5}, None, "Rounding Errors",
                 dda_expected_small, id="DriftDiffusionAnalytical-SmallDriftRate-NegInput"),
    pytest.param(Functions.DriftDiffusionAnalytical, 1e-4,
                 {"drift_rate": -1e-5, "threshold": RAND2, "starting_value": RAND3,
                  "non_decision_time":RAND4, "noise": RAND5}, None, "Rounding Errors",
                 dda_expected_small, id="DriftDiffusionAnalytical-SmallNegDriftRate"),
    # Two tests with different inputs to show that input is ignored.
    pytest.param(Functions.NormalDist, 1e14, {"mean": RAND1, "standard_deviation": RAND2},
                 None, None, normal_expected_mt, id="NormalDist"),
    pytest.param(Functions.NormalDist, 1e-4, {"mean": RAND1, "standard_deviation": RAND2},
                 None, None, normal_expected_mt, id="NormalDist Small Input"),
    pytest.param(Functions.UniformDist, 1e14, {"low": min(RAND1, RAND2), "high": max(RAND1, RAND2)},
                 None, None, uniform_expected_mt, id="UniformDist"),
    pytest.param(Functions.UniformDist, 1e-4, {"low": min(RAND1, RAND2), "high": max(RAND1, RAND2)},
                 None, None, uniform_expected_mt, id="UniformDist"),
    # Inf inputs select Philox PRNG, test_var should never be inf
    pytest.param(Functions.NormalDist, 1e14, {"mean": RAND1, "standard_deviation": RAND2},
                 _SeededPhilox, None, normal_expected_philox, id="NormalDist Philox"),
    pytest.param(Functions.NormalDist, 1e-4, {"mean": RAND1, "standard_deviation": RAND2},
                 _SeededPhilox, None, normal_expected_philox, id="NormalDist Philox"),
    pytest.param(Functions.UniformDist, 1e14, {"low": min(RAND1, RAND2), "high": max(RAND1, RAND2)},
                 _SeededPhilox, None, uniform_expected_philox, id="UniformDist Philox"),
    pytest.param(Functions.UniformDist, 1e-4, {"low": min(RAND1, RAND2), "high": max(RAND1, RAND2)},
                 _SeededPhilox, None, uniform_expected_philox, id="UniformDist Philox"),
]

@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, prng, llvm_skip, expected", test_data)
def test_execute(func, variable, params, prng, llvm_skip, expected, benchmark, func_mode):
    benchmark.group = "TransferFunction " + func.componentName

    if func_mode != 'Python':
        precision = pytest.helpers.llvm_current_fp_precision()
        # PTX needs only one special case, this is not worth adding it to the mechanism above
        if func_mode == "PTX" and precision == 'fp32' and expected is dda_expected_negative:
            expected = (0.4236549735069275, 5.960464477539063e-08,
                        0.5173678398132324, 0.06942932307720184, 6.302994728088379,
                        1.4934064149856567, 0.4288918972015381, 1.7737658023834229)

        expected = llvm_expected.get(precision, {}).get(expected, expected)

    if expected is None:
        pytest.skip(llvm_skip)

    f = func(default_variable=variable, **params)
    if prng is not None:
        f.parameters.random_state.set(prng([0]))

    ex = pytest.helpers.get_func_execution(f, func_mode)
    res = benchmark(ex, variable)

    np.testing.assert_allclose(res, expected)



import pytest

import psyneulink as pnl
import psyneulink.core.components.functions.nonstateful.distributionfunctions
import psyneulink.core.components.functions.stateful.integratorfunctions


class TestNaming:
    # ------------------------------------------------------------------------------------------------

    # NAMING CONVENTIONS

# ------------------------------------------------------------------------------------------------
    # TEST 1
    # Test that Compositions are given default names with incrementing index, starting at 0

    def test_composition_names(self):

        T = pnl.TransferMechanism()
        C1 = pnl.Composition(pathways=[T])
        C2 = pnl.Composition(pathway=[T])
        assert C1.name == 'Composition-0'
        assert C2.name == 'Composition-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 2
    # Test that Processes and Systems assigned duplicate names are indexed starting at 1 (original is not indexed)

    def test_composition_default_names_2(self):

        T = pnl.TransferMechanism(name='T0')
        C1 = pnl.Composition(name='MY COMPOSITION', pathways=[T])
        C2 = pnl.Composition(name='MY COMPOSITION', pathways=[T])
        assert C1.name == 'MY COMPOSITION'
        assert C2.name == 'MY COMPOSITION-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 3
    # Test that Mechanisms are given default names with incrementing index, starting at 0

    def test_default_mechanism_names(self):
        T1 = pnl.TransferMechanism()
        T2 = pnl.TransferMechanism()
        assert T1.name == 'TransferMechanism-0'
        assert T2.name == 'TransferMechanism-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 4
    # Test that Mechanism assigned a duplicate name is incremented starting at 1 (original is not indexed)

    @pytest.mark.parametrize(
        'name, expected_list',
        [
            ('MY TRANSFER MECHANISM', ['MY TRANSFER MECHANISM', 'MY TRANSFER MECHANISM-1']),
            ('A-1', ['A-1', 'A-1-1', 'A-1-2']),
            ('A', ['A', 'A-1', 'A-2']),
        ]
    )
    def test_duplicate_assigned_mechanism_names(self, name, expected_list):
        for expected_name in expected_list:
            t = pnl.TransferMechanism(name=name)
            assert t.name == expected_name

    def test_duplicate_assigned_mechanism_names_2(self):
        pnl.TransferMechanism(name='A')
        pnl.TransferMechanism(name='A')  # A-1
        pnl.TransferMechanism(name='A')  # A-2
        t = pnl.TransferMechanism(name='A-1')

        assert t.name == 'A-3'

    # ------------------------------------------------------------------------------------------------
    # TEST 5
    # Test that default MappingProjections in deferred init are assigned indexed names

    def test_deferred_init_default_MappingProjection_names(self):
        P1 = pnl.MappingProjection()
        P2 = pnl.MappingProjection()
        assert P1.name == 'Deferred Init MappingProjection'
        assert P2.name == 'Deferred Init MappingProjection-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 6
    # Test that MappingProjections with assigned names in deferred init are assigned indexed names

    def test_deferred_init_assigned_MappingProjection_names(self):
        PN1 = pnl.MappingProjection(name='MY PROJECTION')
        PN2 = pnl.MappingProjection(name='MY PROJECTION')
        assert PN1.name == 'MY PROJECTION [Deferred Init]'
        assert PN2.name == 'MY PROJECTION [Deferred Init]-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 7
    # Test that default ModulatoryProjections in deferred init are assigned indexed names

    def test_deferred_init_default_ModulatoryProjection_names(self):
        LP1 = pnl.LearningProjection()
        LP2 = pnl.LearningProjection()
        assert LP1.name == 'Deferred Init LearningProjection'
        assert LP2.name == 'Deferred Init LearningProjection-1'

        CP1 = pnl.ControlProjection()
        CP2 = pnl.ControlProjection()
        assert CP1.name == 'Deferred Init ControlProjection'
        assert CP2.name == 'Deferred Init ControlProjection-1'

        GP1 = pnl.GatingProjection()
        GP2 = pnl.GatingProjection()
        assert GP1.name == 'Deferred Init GatingProjection'
        assert GP2.name == 'Deferred Init GatingProjection-1'

    # ------------------------------------------------------------------------------------------------
    # TEST 8
    # Test that objects of different types can have the same name

    def test_different_object_types_with_same_names(self):
        T1 = pnl.TransferMechanism(name='MY NAME')
        T2 = pnl.TransferMechanism(name='MY NAME')
        P1 = pnl.MappingProjection(sender=T1, receiver=T2, name='MY NAME')
        assert T1.name == 'MY NAME'
        assert P1.name == 'MY NAME'

    # ------------------------------------------------------------------------------------------------
    # TEST 9
    # Test that InputPorts and Projections constructed on their own and assigned are properly named

    def test_input_port_and_assigned_projection_names(self):
        T1 = pnl.TransferMechanism(name='T1')
        T2 = pnl.TransferMechanism(name='T2', input_ports=[T1])
        I1 = pnl.InputPort(owner=T2)
        I2 = pnl.InputPort(projections=[T1])
        assert I2.name == 'Deferred Init InputPort'
        T2.add_ports([I2])
        assert I1.name == 'InputPort-1'
        assert I2.name == 'InputPort-2'
        assert T2.input_ports[0].path_afferents[0].name == \
               'MappingProjection from T1[RESULT] to T2[InputPort-0]'
        assert T2.input_ports[2].path_afferents[0].name == \
               'MappingProjection from T1[RESULT] to T2[InputPort-2]'

    # ------------------------------------------------------------------------------------------------
    # TEST 10
    # Test that OutputPorts are properly named

        T1 = pnl.TransferMechanism(output_ports=['MY OUTPUT_PORT',[0]])
        assert T1.output_ports[0].name == 'MY OUTPUT_PORT'
        assert T1.output_ports[1].name == 'OutputPort-0'
        O = pnl.OutputPort(owner=T1)
        assert T1.output_ports[2].name == 'OutputPort-1'
        O2 = pnl.OutputPort()
        T1.add_ports([O2])
        assert T1.output_ports[3].name == 'OutputPort-2'

    # ------------------------------------------------------------------------------------------------
    # TEST 11
    # Test that ControlSignals and ControlProjections are properly named
    @pytest.mark.control
    @pytest.mark.parametrize('control_spec', [pnl.CONTROL, pnl.PROJECTIONS])
    def test_control_signal_and_control_projection_names(self, control_spec):
        D1 = pnl.DDM(name='D1')
        D2 = pnl.DDM(name='D2')

        # ControlSignal with one ControlProjection
        C1 = pnl.ControlMechanism(control_signals=[D1.parameter_ports[
                                                       psyneulink.core.components.functions.nonstateful.distributionfunctions.DRIFT_RATE]])
        assert C1.control_signals[0].name == 'D1[drift_rate] ControlSignal'
        assert C1.control_signals[0].efferents[0].name == 'ControlProjection for D1[drift_rate]'

        # ControlSignal with two ControlProjection to two parameters of same Mechanism
        C2 = pnl.ControlMechanism(control_signals=[{control_spec:[D1.parameter_ports[
                                                                      psyneulink.core.components.functions.nonstateful.distributionfunctions.DRIFT_RATE],
                                                                  D1.parameter_ports[
                                                                      psyneulink.core.globals.keywords.THRESHOLD]]}])
        assert C2.control_signals[0].name == 'D1[drift_rate, threshold] ControlSignal'
        assert C2.control_signals[0].efferents[0].name == 'ControlProjection for D1[drift_rate]'
        assert C2.control_signals[0].efferents[1].name == 'ControlProjection for D1[threshold]'

        # ControlSignal with two ControlProjection to two parameters of different Mechanisms
        C3 = pnl.ControlMechanism(control_signals=[{control_spec:[D1.parameter_ports[
                                                                         psyneulink.core.components.functions.nonstateful.distributionfunctions.DRIFT_RATE],
                                                                     D2.parameter_ports[
                                                                         psyneulink.core.components.functions.nonstateful.distributionfunctions.DRIFT_RATE]]}])
        assert C3.control_signals[0].name == 'ControlSignal-0 divergent ControlSignal'
        assert C3.control_signals[0].efferents[0].name == 'ControlProjection for D1[drift_rate]'
        assert C3.control_signals[0].efferents[1].name == 'ControlProjection for D2[drift_rate]'

    # ------------------------------------------------------------------------------------------------
    # TEST 12
    # Test that GatingSignals and GatingProjections are properly named

    @pytest.mark.control
    def test_gating_signal_and_gating_projection_names(self):
        T3 = pnl.TransferMechanism(name='T3')
        T4 = pnl.TransferMechanism(name='T4', input_ports=['First Port','Second Port'])

        # GatingSignal with one GatingProjection
        G1 = pnl.GatingMechanism(gating_signals=[T3])
        assert G1.gating_signals[0].name == 'T3[InputPort-0] GatingSignal'
        assert G1.gating_signals[0].efferents[0].name == 'GatingProjection for T3[InputPort-0]'

        # GatingSignal with two GatingProjections to two Ports of same Mechanism
        G2 = pnl.GatingMechanism(gating_signals=[{pnl.PROJECTIONS:[T4.input_ports[0], T4.input_ports[1]]}])
        assert G2.gating_signals[0].name == 'T4[First Port, Second Port] GatingSignal'
        assert G2.gating_signals[0].efferents[0].name == 'GatingProjection for T4[First Port]'
        assert G2.gating_signals[0].efferents[1].name == 'GatingProjection for T4[Second Port]'

        # GatingSignal with two GatingProjections to two Ports of different Mechanisms
        G3 = pnl.GatingMechanism(gating_signals=[{pnl.PROJECTIONS:[T3, T4]}])
        assert G3.gating_signals[0].name == 'GatingSignal-0 divergent GatingSignal'
        assert G3.gating_signals[0].efferents[0].name == 'GatingProjection for T3[InputPort-0]'
        assert G3.gating_signals[0].efferents[1].name == 'GatingProjection for T4[First Port]'

        # GatingProjections to ProcessingMechanism from GatingSignals of existing GatingMechanism
        T5 = pnl.TransferMechanism(name='T5',
                                   input_ports=[T3.output_ports[pnl.RESULT],
                                                 G3.gating_signals['GatingSignal-0 divergent GatingSignal']],
                                   output_ports=[G3.gating_signals['GatingSignal-0 divergent GatingSignal']])

    def test_composition_names(self):
        C1 = pnl.Composition()
        C2 = pnl.Composition()
        C3 = pnl.Composition()

        assert C1.name == 'Composition-0'
        assert C2.name == 'Composition-1'
        assert C3.name == 'Composition-2'



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM bindings **************************************************************

from llvmlite import binding
import time
import warnings

from .builder_context import LLVMBuilderContext, _find_llvm_function, _gen_cuda_kernel_wrapper_module
from .builtins import _generate_cpu_builtins_module
from .debug import debug_env

try:
    import pycuda
    # Do not continue if the version is too old
    if pycuda.VERSION[0] < 2018:
        raise UserWarning("pycuda too old (need 2018+): " + str(pycuda.VERSION))
    import pycuda.driver
    # pyCUDA needs to be built against 6+ to enable Linker
    if pycuda.driver.get_version()[0] < 6:
        raise UserWarning("CUDA driver too old (need 6+): " + str(pycuda.driver.get_version()))

    from pycuda import autoinit as pycuda_default
    import pycuda.compiler
    assert pycuda_default.context is not None
    pycuda_default.context.set_cache_config(pycuda.driver.func_cache.PREFER_L1)
    ptx_enabled = True
    if "cuda-check" in debug_env:
        print("PsyNeuLink: CUDA backend enabled!")
except Exception as e:
    if "cuda-check" in debug_env:
        warnings.warn("Failed to enable CUDA/PTX: {}".format(e))
    ptx_enabled = False


__all__ = ['cpu_jit_engine', 'ptx_enabled']

if ptx_enabled:
    __all__.append('ptx_jit_engine')


# Compiler binding
__initialized = False


def _binding_initialize():
    global __initialized
    if not __initialized:
        binding.initialize()
        if not ptx_enabled:
            # native == currently running CPU. ASM printer includes opcode emission
            binding.initialize_native_target()
            binding.initialize_native_asmprinter()
        else:
            binding.initialize_all_targets()
            binding.initialize_all_asmprinters()

        __initialized = True


def _cpu_jit_constructor():
    _binding_initialize()

    opt_level = int(debug_env.get('opt', 2))

    # PassManagerBuilder can be shared
    __pass_manager_builder = binding.PassManagerBuilder()
    __pass_manager_builder.loop_vectorize = opt_level != 0
    __pass_manager_builder.slp_vectorize = opt_level != 0
    __pass_manager_builder.opt_level = opt_level

    __cpu_features = binding.get_host_cpu_features().flatten()
    __cpu_name = binding.get_host_cpu_name()

    # Create compilation target, use default triple
    __cpu_target = binding.Target.from_default_triple()
    # FIXME: reloc='static' is needed to avoid crashes on win64
    # see: https://github.com/numba/llvmlite/issues/457
    __cpu_target_machine = __cpu_target.create_target_machine(cpu=__cpu_name, features=__cpu_features, opt=opt_level, reloc='static')

    __cpu_pass_manager = binding.ModulePassManager()
    __cpu_target_machine.add_analysis_passes(__cpu_pass_manager)
    __pass_manager_builder.populate(__cpu_pass_manager)

    # And an execution engine with a builtins backing module
    builtins_module = _generate_cpu_builtins_module(LLVMBuilderContext.get_current().float_ty)
    if "dump-llvm-gen" in debug_env:
        with open(builtins_module.name + '.generated.ll', 'w') as dump_file:
            dump_file.write(str(builtins_module))

    __backing_mod = binding.parse_assembly(str(builtins_module))

    __cpu_jit_engine = binding.create_mcjit_compiler(__backing_mod, __cpu_target_machine)
    return __cpu_jit_engine, __cpu_pass_manager, __cpu_target_machine


def _ptx_jit_constructor():
    _binding_initialize()

    opt_level = int(debug_env.get('opt', 2))

    # PassManagerBuilder is used only for inlining simple functions
    __pass_manager_builder = binding.PassManagerBuilder()
    __pass_manager_builder.opt_level = 2
    __pass_manager_builder.size_level = 1

    # The threshold of '64' is empirically selected on GF 3050
    __pass_manager_builder.inlining_threshold = 64

    # Use default device
    # TODO: Add support for multiple devices
    __compute_capability = pycuda_default.device.compute_capability()
    __ptx_sm = "sm_{}{}".format(__compute_capability[0], __compute_capability[1])

    # Create compilation target, use 64bit triple
    __ptx_target = binding.Target.from_triple("nvptx64-nvidia-cuda")
    __ptx_target_machine = __ptx_target.create_target_machine(cpu=__ptx_sm, opt=opt_level)

    __ptx_pass_manager = binding.ModulePassManager()
    __ptx_target_machine.add_analysis_passes(__ptx_pass_manager)
    __pass_manager_builder.populate(__ptx_pass_manager)

    return __ptx_pass_manager, __ptx_target_machine


def _try_parse_module(module):
    module_text_ir = str(module)

    if "dump-llvm-gen" in debug_env:
        with open(module.name + '.generated.ll', 'w') as dump_file:
            dump_file.write(module_text_ir)

    # IR module is not the same as binding module.
    # "assembly" in this case is LLVM IR assembly.
    # This is intentional design decision to ease
    # compatibility between LLVM versions.
    try:
        mod = binding.parse_assembly(module_text_ir)
        mod.verify()
    except Exception as e:
        print("ERROR: llvm parsing failed: {}".format(e))
        mod = None

    return mod


class jit_engine:
    def __init__(self):
        self._jit_engine = None
        self._jit_pass_manager = None
        self._target_machine = None
        self.__mod = None
        # Add an extra reference to make sure it's not destroyed before
        # instances of jit_engine
        self.__debug_env = debug_env

        self.staged_modules = set()
        self.compiled_modules = set()

        # Track few statistics:
        self.__optimized_modules = 0
        self.__linked_modules = 0
        self.__parsed_modules = 0

    def __del__(self):
        if "stat" in self.__debug_env:
            s = type(self).__name__
            print("Total optimized modules in '{}': {}".format(s, self.__optimized_modules))
            print("Total linked modules in '{}': {}".format(s, self.__linked_modules))
            print("Total parsed modules in '{}': {}".format(s, self.__parsed_modules))

    def opt_and_add_bin_module(self, module):
        start = time.perf_counter()
        self._pass_manager.run(module)
        finish = time.perf_counter()

        if "time_stat" in debug_env:
            print("Time to optimize LLVM module bundle '{}': {}".format(module.name, finish - start))

        if "dump-llvm-opt" in self.__debug_env:
            with open(self.__class__.__name__ + '-' + str(self.__optimized_modules) + '.opt.ll', 'w') as dump_file:
                dump_file.write(str(module))

        # This prints generated x86 assembly
        if "dump-asm" in self.__debug_env:
            with open(self.__class__.__name__ + '-' + str(self.__optimized_modules) + '.S', 'w') as dump_file:
                dump_file.write(self._target_machine.emit_assembly(module))

        start = time.perf_counter()
        self._engine.add_module(module)
        self._engine.finalize_object()
        finish = time.perf_counter()
        if "time_stat" in debug_env:
            print("Time to finalize LLVM module bundle '{}': {}".format(module.name, finish - start))
        self.__optimized_modules += 1

    def _remove_bin_module(self, module):
        if module is not None:
            self._engine.remove_module(module)

    def opt_and_append_bin_module(self, module):
        mod_name = module.name
        if self.__mod is None:
            self.__mod = module
        else:
            self._remove_bin_module(self.__mod)
            # Linking here invalidates 'module'
            self.__mod.link_in(module)
            self.__linked_modules += 1

        if "dump-llvm-gen" in debug_env:
            with open(mod_name + '.linked.ll', 'w') as dump_file:
                dump_file.write(str(self.__mod))

        self.opt_and_add_bin_module(self.__mod)

    def clean_module(self):
        self._remove_bin_module(self.__mod)
        self.__mod = None

    @property
    def _engine(self):
        if self._jit_engine is None:
            self._init()

        return self._jit_engine

    @property
    def _pass_manager(self):
        if self._jit_pass_manager is None:
            self._init()

        return self._jit_pass_manager

    def stage_compilation(self, modules):
        self.staged_modules |= modules

    # Unfortunately, this needs to be done for every jit_engine.
    # Liking step in opt_and_add_bin_module invalidates 'mod_bundle',
    # so it can't be linked mutliple times (in multiple engines).
    def compile_staged(self):
        # Parse generated modules and link them
        mod_bundle = binding.parse_assembly("")
        while self.staged_modules:
            m = self.staged_modules.pop()

            start = time.perf_counter()
            new_mod = _try_parse_module(m)
            finish = time.perf_counter()

            if "time_stat" in debug_env:
                print("Time to parse LLVM modules '{}': {}".format(m.name, finish - start))

            self.__parsed_modules += 1
            if new_mod is not None:
                mod_bundle.link_in(new_mod)
                mod_bundle.name = m.name  # Set the name of the last module
                self.compiled_modules.add(m)

        self.opt_and_append_bin_module(mod_bundle)


class cpu_jit_engine(jit_engine):

    def __init__(self, object_cache=None):
        super().__init__()
        self._object_cache = object_cache

    def _init(self):
        assert self._jit_engine is None
        assert self._jit_pass_manager is None
        assert self._target_machine is None

        self._jit_engine, self._jit_pass_manager, self._target_machine = _cpu_jit_constructor()
        if self._object_cache is not None:
            self._jit_engine.set_object_cache(self._object_cache)


# FIXME: Get device side printf pointer
_ptx_builtin_source = """
__device__ {type} __pnl_builtin_sin({type} a) {{ return sin(a); }}
__device__ {type} __pnl_builtin_cos({type} a) {{ return cos(a); }}
__device__ {type} __pnl_builtin_log({type} a) {{ return log(a); }}
__device__ {type} __pnl_builtin_exp({type} a) {{ return exp(a); }}
__device__ {type} __pnl_builtin_pow({type} a, {type} b) {{ return pow(a, b); }}
__device__ int64_t __pnl_builtin_get_printf_address() {{ return 0; }}
"""


class ptx_jit_engine(jit_engine):
    class cuda_engine():
        def __init__(self, tm):
            self._modules = {}
            self._target_machine = tm

            # -dc option tells the compiler that the code will be used for linking
            self._generated_builtins = pycuda.compiler.compile(_ptx_builtin_source.format(type=str(LLVMBuilderContext.get_current().float_ty)), target='cubin', options=['-dc'])

        def set_object_cache(cache):
            pass

        def add_module(self, module):
            max_regs = int(debug_env.get("cuda_max_regs", 256))
            try:
                # LLVM can't produce CUBIN for some reason
                start_time = time.perf_counter()
                ptx = self._target_machine.emit_assembly(module)
                ptx_time = time.perf_counter()
                mod = pycuda.compiler.DynamicModule(link_options=[(pycuda.driver.jit_option.MAX_REGISTERS, max_regs)])
                mod.add_data(self._generated_builtins, pycuda.driver.jit_input_type.CUBIN, "builtins.cubin")
                mod.add_data(ptx.encode(), pycuda.driver.jit_input_type.PTX, module.name + ".ptx")
                module_time = time.perf_counter()
                ptx_mod = mod.link()
                finish_time = time.perf_counter()
                if "time_stat" in debug_env:
                    print("Time to emit PTX module bundle '{}'({} lines): {}".format(module.name, len(ptx.splitlines()), ptx_time - start_time))
                    print("Time to add PTX module bundle '{}': {}".format(module.name, module_time - ptx_time))
                    print("Time to link PTX module bundle '{}': {}".format(module.name, finish_time - module_time))
                    print("Total time to process PTX module bundle '{}': {}".format(module.name, finish_time - start_time))

            except Exception as e:
                print("FAILED to generate PTX module:", e)
                print(ptx)
                return None

            self._modules[module] = ptx_mod

        def finalize_object(self):
            pass

        def remove_module(self, module):
            self._modules.pop(module, None)

        def _find_kernel(self, name):
            function = None
            for m in self._modules.values():
                try:
                    function = m.get_function(name)
                except pycuda._driver.LogicError:
                    pass
            return function

    def __init__(self, object_cache=None):
        super().__init__()
        self._object_cache = object_cache

    def _init(self):
        assert self._jit_engine is None
        assert self._jit_pass_manager is None
        assert self._target_machine is None

        self._jit_pass_manager, self._target_machine = _ptx_jit_constructor()
        self._jit_engine = ptx_jit_engine.cuda_engine(self._target_machine)

    def get_kernel(self, name):
        kernel = self._engine._find_kernel(name + "_cuda_kernel")
        if kernel is None:
            function = _find_llvm_function(name)
            wrapper_mod = _gen_cuda_kernel_wrapper_module(function)
            self.stage_compilation({wrapper_mod})
            self.compile_staged()
            kernel = self._engine._find_kernel(name + "_cuda_kernel")
#            kernel.set_cache_config(pycuda.driver.func_cache.PREFER_L1)

        return kernel



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  TimerMechanism *************************************************

"""

Contents
--------

  * `TimerMechanism_Overview`
  * `TimerMechanism_Creation`
  * `TimerMechanism_Structure`
  * `TimerMechanism_Execution`
  * `TimerMechanism_Examples`
  * `TimerMechanism_Class_Reference`


.. _TimerMechanism_Overview:

Overview
--------

A TimerMechanism is a type of `IntegratorMechanism` the `value <Mechanism_Base.value>` of which begins at a specified
`start <TimerMechanism.start>` value, is changed monotonically each time it is executed, until it reaches a specified
`end <TimerMechanism.end>` value.  The number of executions it takes to do so is determined by a combination of its
`duration <TimerMechanism.duration>` and `increment <TimerMechanism.increment>` parameters, and whether or not it
receives any input; and the path that its `value <Mechanism_Base.value>` takes is determined by its `trajectory
<TimerMechanism.trajectory>` Function. It can be reset to its starting value by calling its `reset
<TimerMechanism.reset>` method, or by modulating its `reset <TimerMechanism.reset>` Parameter with a `ControlSignal`.

A TimerMechanism can be used to implement a variety of time-based processes, such as the collapse of a boundary over
time, or the rise of a value to a threshold. It can also be configured to execute multiple such processes in parallel,
each with its own starting and ending values, as well as direction, increment and input (on a given execution), all of
which use the same `trajectory <TimerMechanism.trajectory>` Function.

.. _TimerMechanism_Creation:

Creating a TimerMechanism
-------------------------

A TimerMechanism can be created directly by calling its constructor, or using the `mechanism` command and specifying
*TIMER_MECHANISM* as its **mech_spec** argument. It can be created with or without a source of input. By default, a
TimerMechanisms increments linearly, starting at 0, incrementing by 0.01 each time it is executed, and stopping when it
reaches 1. However, the shape, starting, ending and rate of increment can all be configured.  The shape of the timer's
progression is specified by it **trajectory** argument, which must be a `TimerFunction` or an appropriately configured
`UserDefinedFunction` (see `below <TimerMechanism_Trajectory_Function>` for details); the starting and ending `values
<Mechanism_Base.value>` of the timer are specified by its **start** and **end** arguments, respectively, and the ammount
it progresses each time the Mechanimsm is executed (in the absence of input) is specified by its **increment** argument.

.. _TimerMechanism_Structure:

Structure
---------

A TimerMechanism may or may not have a source of input, and has a single `OutputPort`.  Its `function
<TimerMechanism.function>` is an `SimpleIntegrator` `Function` that determines the input to the TimerMechanism's
`trajectory <TimerMechanism.trajectory>` Function which, in turn, determines the `value <Mechanism_Base.value>` of the
TimerMechanism. The TimerMechanism's `function <TimerMechanism.function>` is always a `SimpleIntegrator` that always
starts at 0 and, each time it is executed, increments the value of the TimerMechanism based either on its input or, if
it receives none, then its `increment <TimerMechanism.increment>` Parameter (see `TimerMechanism_Execution` for
additional details).

    .. technical_note::
       The TimerMechanism's `function <TimerMechanism.function>` cannot be modified; its `rate
       <SimpleIntegrator.rate>` is set to the value specified for the TimerMechanism's `increment
       <TimerMechanism.increment>` Parameter, and its `initializer <SimpleIntegrator.initializer>`,
       `noise <SimpleIntegrator.noise>` and `offset <SimpleIntegrator.offset>` Parameters are fixed at 0

.. _TimerMechanism_Trajectory_Function:

Its `trajectory <TimerMechanism.trajectory>` Function must be either a `TimerFunction` or a `UserDefinedFunction` that
has **initial**, **final** and *duration** parameters, and takes a single numeric value, a list or an np.array as its
`variable <Function_Base.variable>`.

    .. technical_note::
       The `trajectory <TimerMechanism.trajectory>` Function is an auxilliary function that takes the result of the
       TimerMechanism's `function <TimerMechanism.function>` and transforms it to implement the specified trajectory
       of the timer's `value <Mechanism_Base.value>`. The value of the `start <TimerMechanism.start>` parameter is
       assigned as the `initial <TimerFunction.initial>` Parameter of its `trajectory <TimerMechanism.trajectory>`
       Function; its `end <TimerMechanism.end>` Parameter is assigned as the `final <TimerFunction.final>` Parameter of
       the `trajectory <TimerMechanism.trajectory>` Function, and its `duration <TimerMechanism.duration>` Parameter
       is assigned as the `duration <TimerFunction.duration>` Parameter of the `trajectory <TimerMechanism.trajectory>`;
       which is also used to set the TimerMechanism's `finished <TimeMechanism.finished>` attribute.

.. _TimerMechanism_Modulation:

A TimerMechanism's `start <TimerMechanism.start>`, `end <TimerMechanism.end>` and `increment <TimerMechanism.increment>`
parameters can be `modulated <ModulatorySignal_Modulation>` by a `ControlMechanism`.

.. _TimerMechanism_Execution:

Execution
---------

When a TimerMechanism is executed, it advances its value by adding either its `input <TimerMechanism.input>`
or its `increment <TimerMechanism.increment>` Parameter to the `previous_value <IntegratorFunction.previous_value>`
of its `function <TimerMechanism.function>`, that is then passed to the `Function` specified by its `trajectory
<TimerMechanism.trajectory>` Parameter, the result of which is assigned to the `value <Mechanism_Base.value>` of the
TimerMechansim's `primary OutputPort <OutputPort_Primary>`. If its `default_variable <TimerMechanism.default_variable>`
is a scalar or of length 1, or  its **input_shapes** is specified as 1, the TimerMechanism generates a scalar value
as its output; otherwise, it generates an array of values, each element of which is independently advanced. If its
`increment <TimerMechanism.increment>` Parameter is a single value, that is used for advancing all elements; if the
`increment <TimerMechanism.increment>` Parameter is a list or array, then each element is used to increment the
corresponding element of the timer array.

The TimerMechanism stops advancing after the `value <Function_Base.value>` of its `function <TimerMechanism.function>`
equals the TimerMechanism's `duration <TimerMechanism.duration>` Parameter.  If the TimerMechanism receives no input,
then it will stop advancing after the number of executions = `duration <TimerMechanism.duration>` / `increment
<TimerMechanism.increment>`.  If the TimerMechanism receives input, then the number of executions is determined by the
number of times it is executed, and the amount of input it receives on each execution.  When the TimerMechanism stops
advancing, it sets its `finished <TimerMechanism.finished>` attribute to `True`, and its `value <Mechanism_Base.value>`
remains equal to its `end <TimerMechanism.end>` on any further executions, unless it is reset.  If the TimerMechanism
is reset, its `value <Mechanism_Base.value>` is set to its `start <TimerMechanism.start>` value.

    .. hint::
       A TimerMechanism's `finished <TimerMechanism.finished>` attribute can be used together with a `Scheduler`
       `Condition` to make the processing of other `Mechanisms <Mechanism>` contingent on the TimerMechanism
       reaching its end value.

    .. note::
       The TimerMechanism's `finished <TimerMechanism.finished>` attribute is not used to determine when it stops
       advancing; rather, it is set to `True` when the TimerMechanism's `value <Mechanism_Base.value>` reaches its
       `end <TimerMechanism.end>` value.

If a TimerMechanism continues to be executed after its `finished <TimerMechanism.finished>` attribute is True, its
`value <Mechanism_Base.value>` remains equal to its `end <TimerMechanism.end>` value.  The TimerMechanism can be reset
to its `start <TimerMechanism.start>` value by setting its `reset <TimerMechanism.reset>` Parameter to a non-zero value,
as described for a standard `IntegratorMechanism <IntegratorMechanism_Reset>`.

.. _TimerMechanism_Examples:

Examples
--------

The following example creates a TimerMechanism with a linear decay from 1 to 0:

    >>> import psyneulink as pnl
    >>> my_timer_mechanism = pnl.TimerMechanism(trajectory=LINEAR, start=1, end=0))


.. _TimerMechanism_Class_Reference:

Class Reference
---------------

"""
from collections.abc import Iterable

from beartype import beartype

from psyneulink._typing import Optional, Union
import numpy as np

from psyneulink.core.components.functions.nonstateful.timerfunctions import TimerFunction, LinearTimer
from psyneulink.core.components.functions.stateful.integratorfunctions import IntegratorFunction, SimpleIntegrator
from psyneulink.core.components.mechanisms.processing.integratormechanism import IntegratorMechanism
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.globals.keywords import TIMER_MECHANISM, PREFERENCE_SET_NAME, RESET, TRAJECTORY
from psyneulink.core.globals.parameters import Parameter, FunctionParameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = [
    'DEFAULT_RATE', 'TimerMechanism', 'TimerMechanismError'
]

# TimerMechanism parameter keywords:
DEFAULT_RATE = 1


class TimerMechanismError(MechanismError):
    pass


class TimerMechanism(IntegratorMechanism):
    """
    TimerMechanism(                             \
        start=0,                                \
        increment=1,                            \
        end=1,                                  \
        function=SimpleIntegrator(rate=1),      \
        trajectory=LinearTimer,                 \
        duration=1)

    Subclass of `IntegratorMechanism` that advances its input until it reaches a specified value.
    See `IntegratorMechanism <IntegratorMechanism_Class_Reference>` for additional arguments and attributes.

    Arguments
    ---------

    start : scalar, list or array : default 0
        specifies the starting `value <Mechanism_Base.value>` of the timer; if a list or array, the length must be
        the same as specified for **default_variable** or **input_shapes** (see `TimerMechanism_Execution` for
        additional details).

    increment : scalar, list or array : default 1
        specifies the amount by which the `previous_value <IntegratorFunction.previous_value>` of the
        TimerMechanism's `function <TimerMechanism.function>` is incremented each time the TimerMechanism is
        executed in the absence of input; if a list or array, the length must be the same as specified for
        **default_variable** or **input_shapes** (see `TimerMechanism_Execution` for additional details).

    function : IntegratorFunction : default SimpleIntegrator(rate=1)
        specifies the function used to increment the input to the TimerMechanism's `trajectory
        <TimerMechanism.trajectory>` Function; must take a single numeric value, or a list or np.array
        of values, and return one of the same shape.

    trajectory : TransferFunction or UserDefinedFunction : default LinearTimer
        specifies the shape of the timer's trajectory; must be a supported `TransferFunction`
        (see `trajectory function <TimerMechanism_Trajectory_Function>`)

    end : scalar, list or array : default 1
        specifies the value of its `trajectory <TimerMechamism.trajectory>` function at which the timer stops advancing;
        if a list or array, the length must be the same as specified for **default_variable** or **input_shapes** (see
        `TimerMechanism_Execution` for additional details).

    duration : scalar, list or array : default 1
        specifies the value of its `variable <Mechanism_Base.variable>` at which the timer stops advancing; if a list
        or array, the length must be the same as specified for **default_variable** or **input_shapes** (see
        `TimerMechanism_Execution` for additional details).

    reset_default : number, list or np.ndarray : default 0
        specifies the default value used for the `reset <TimerMechanism.reset>` parameter.

    Attributes
    ----------

    start : scalar, list or array
        determines the starting `value <Mechanism_Base.value>` of the timer; assigned as the `start` Parameter of the
        TimerMechanism's `trajectory <TimerMechanism.function>` if it has one; otherwise, it is computed if possible
        or, an error is raised (see `TimerMechanism_Execution` for additional details).

    increment : scalar, list or array
        determines the amount by which the `previous_value <IntegratorFunction.previous_value>` of the
        TimerMechanism's `function <TimerMechanism.function>` is incremented each time the TimerMechanism is
        executed in the absence of input; assigned as the `rate <IntegratorFunction.rate>` of the TimerMechanism's
        `function <TimerMechanism.function>` (see `TimerMechanism_Execution` for additional details).

    function : IntegratorFunction
        determines the function used to advance the input to the TimerMechanism's `trajectorytrajectory
        <TimerMechanism.trajectory>` Function; if the TimerMechanism receives an external input, that is
        used to advance the timer; otherwise, the `increment <TimerMechanism.increment>` Parameter is used.

    trajectory : TimerFunction or UserDefinedFunction
        determines the `Function` used to transform the ouput of the TimerMechanism's `function
        <TimerMechanism.function>` to generate its output; this determines the shape of the trajectory
        of the TimerMechanism's `value <Mechanism_Base.value>`.

    end : scalar, list or array : default 1
        determines the value of its `trajectory <TimerMechamism.trajectory>` function at which the timer stops
        advancing; if a list or array, the length must be the same as specified for **default_variable** or
        **input_shapes** (see `TimerMechanism_Execution` for additional details).

    duration : scalar, list or array
        determines the value at which the timer stops advancing, after which it sets its `finished
        <TimerMechanism.finished>` Parameter to `True` and `value <Mechanism_Base.value>` remains equal to
        its `duration <TimerMechanism.duration>` value.

    finished : bool
        indicates whether the TimerMechanism has reached its `duration <TimerMechanism.duration>` value
        (see `TimerMechanism_Execution` for additional details).

    reset : int, float or 1d array of length 1 : default 0
        if non-zero, the TimerMechanism's `reset <Mechanism_Base.reset>` method is called, which resets the
        `value <Mechanism_Base.value>` of the TimerMechanism to its initial value (see
        `TimerMechanism_Reset` for additional details).

    """
    componentType = TIMER_MECHANISM

    classPreferenceLevel = PreferenceLevel.TYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'TimerMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(IntegratorMechanism.Parameters):
        """
            Attributes
            ----------

                finished
                    see `finished <TimerMechanism.finished>`

                    :default value: False
                    :type: `bool`

                duration
                    see `duration <TimerMechanism.duration>`

                    :default value: 1
                    :type: `float`

                function
                    see `function <TimerMechanism.function>`

                    :default value: `SimpleIntegrator`(initializer=numpy.array([0]), rate=1)
                    :type: `Function`

                increment
                    see `increment <TimerMechanism.increment>`

                    :default value: 1
                    :type: `float`

                start
                    see `start <TimerMechanism.start>`

                    :default value: 0
                    :type: `float`

                end
                    see `end <TimerMechanism.end>`

                    :default value: 1
                    :type: `float`

                trajectory
                    see `trajectory <TimerMechanism.trajectory>`

                    :default value: `LinearTimer`
                    :type: `Function`
        """
        function = Parameter(SimpleIntegrator, stateful=False, loggable=False)
        trajectory = Parameter(LinearTimer, stateful=False, loggable=False)
        start = FunctionParameter(0, function_name='trajectory', function_parameter_name='initial', primary=True)
        increment = FunctionParameter(.01, function_name='function', function_parameter_name='rate', primary=True)
        end = FunctionParameter(1, function_name='trajectory', function_parameter_name='final', primary=True )
        duration = FunctionParameter(1, function_name='trajectory', function_parameter_name='duration', primary=True )
        finished = Parameter(False, stateful=True, loggable=True)

        def _validate_trajectory(self, trajectory):
            if not (isinstance(trajectory, TimerFunction)
                    or (isinstance(trajectory, type) and issubclass(trajectory, TimerFunction))):
                return f'must be a TimerFunction'

        def _validate_start(self, start):
            if not isinstance(start, (int, float, list, np.ndarray)):
                return f'must be an int, float or a list or array of either'

        def _validate_increment(self, increment):
            if not isinstance(increment, (int, float, list, np.ndarray)):
                return f'must be an int, float or a list or array of either'

        def _validate_end(self, end):
            if not isinstance(end, (int, float, list, np.ndarray)):
                return f'must be an int, float or a list or array of either'

        def _validate_durat(self, duration):
            if not isinstance(duration, (int, float, list, np.ndarray)):
                return f'must be an int, float or a list or array of either'

    @check_user_specified
    @beartype
    def __init__(self,
                 input_shapes=None,
                 start=None,
                 increment=None,
                 function=None,
                 trajectory=None,
                 end=None,
                 duration=None,
                 params=None,
                 name=None,
                 prefs: Optional[ValidPrefSet] = None,
                 **kwargs):
        """Assign type-level preferences, default input value (SigmoidLayer_DEFAULT_BIAS) and call super.__init__
        """

        super(TimerMechanism, self).__init__(default_variable=1,
                                             input_shapes=input_shapes,
                                             start=start,
                                             increment=increment,
                                             function=function,
                                             trajectory=trajectory,
                                             end=end,
                                             duration=duration,
                                             finished=False,
                                             params=params,
                                             name=name,
                                             prefs=prefs,
                                             **kwargs)

    def _execute(self, variable=None, context=None, runtime_params=None, **kwargs):
        """Override to check for call to reset by ControlSignal"""
        if not self.is_initializing and self.parameters.finished.get(context):
            return self.trajectory(self.function.parameters.previous_value._get(context))

        x = super()._execute(variable=variable, context=context, runtime_params=runtime_params, **kwargs)
        y = self.trajectory(x)

        # No need to reset during initialization (which will occur if **reset_default** != 0)
        if not self.is_initializing:

            if np.allclose(x,self.parameters.duration.get(context)):
                self.parameters.finished._set(True, context)

            # If reset Parameter has been set, reset() TimerMechanism to start value
            if np.array(self._get_current_parameter_value(RESET,context)).squeeze():
                self.reset(self.parameters.start._get(context), context=context)
                y = self.parameters.value._get(context).reshape(y.shape)

        return y

    def reset(self, *args, force=False, context=None, **kwargs):
        super().reset(*args, force=force, context=context, **kwargs)
        self.finished = False



import numpy as np

from psyneulink.core.components.functions.stateful.integratorfunctions import FitzHughNagumoIntegrator
from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear
from psyneulink.core.components.mechanisms.processing.integratormechanism import IntegratorMechanism
from psyneulink.core.compositions.composition import Composition
from psyneulink.library.components.mechanisms.processing.transfer.lcamechanism import LCAMechanism


class TestGilzenratMechanisms:

    def test_defaults(self):
        G = LCAMechanism(integrator_mode=True,
                         leak=1.0,
                         noise=0.0,
                         time_step_size=0.02,
                         function=Linear,
                         self_excitation=1.0,
                         competition=-1.0)

        # - - - - - LCAMechanism integrator functions - - - - -
        # X = previous_value + (rate * previous_value + variable) * self.time_step_size + noise
        # f(X) = 1.0*X + 0

        np.testing.assert_allclose(G.execute(), np.array([[0.0]]))
        # X = 0.0 + (0.0 + 0.0)*0.02 + 0.0
        # X = 0.0 <--- previous value 0.0
        # f(X) = 1.0*0.0  <--- return 0.0, recurrent projection 0.0

        np.testing.assert_allclose(G.execute(1.0), np.array([[0.02]]))
        # X = 0.0 + (0.0 + 1.0)*0.02 + 0.0
        # X = 0.02 <--- previous value 0.02
        # f(X) = 1.0*0.02  <--- return 0.02, recurrent projection 0.02

        # Outside of a Composition, previous value works (integrator) but recurrent projection does NOT
        np.testing.assert_allclose(G.execute(1.0), np.array([[0.0396]]))
        # X = 0.02 + (-0.02 + 1.0)*0.02 + 0.0
        # X = 0.0396 --- previous value 0.0396
        # f(X) = 1.0*0.0396 <--- return 0.02, recurrent projection 0.02

    def test_previous_value_stored(self):
        G = LCAMechanism(integrator_mode=True,
                         leak=1.0,
                         noise=0.0,
                         time_step_size=0.02,
                         function=Linear(slope=2.0),
                         self_excitation=1.0,
                         competition=-1.0,
                         initial_value=np.array([[1.0]]))

        C = Composition(pathways=[G])
        G.output_port.parameters.value.set([0.0], override=True)

        # - - - - - LCAMechanism integrator functions - - - - -
        # X = previous_value + (rate * previous_value + variable) * self.time_step_size + noise
        # f(X) = 2.0*X + 0

        # - - - - - starting values - - - - -
        # variable = G.output_port.value + stimulus = 0.0 + 1.0 = 1.0
        # previous_value = initial_value = 1.0
        # single_run = S.execute([[1.0]])
        # np.testing.assert_allclose(single_run, np.array([[2.0]]))
        np.testing.assert_allclose(C.execute(inputs={G:[[1.0]]}), np.array([[2.0]]))
        # X = 1.0 + (-1.0 + 1.0)*0.02 + 0.0
        # X = 1.0 + 0.0 + 0.0 = 1.0 <--- previous value 1.0
        # f(X) = 2.0*1.0  <--- return 2.0, recurrent projection 2.0

        np.testing.assert_allclose(C.execute(inputs={G:[[1.0]]}), np.array([[2.08]]))
        # X = 1.0 + (-1.0 + 3.0)*0.02 + 0.0
        # X = 1.0 + 0.04 = 1.04 <--- previous value 1.04
        # f(X) = 2.0*1.04  <--- return 2.08

        np.testing.assert_allclose(C.execute(inputs={G:[[1.0]]}), np.array([[2.1616]]))
        # X = 1.04 + (-1.04 + 3.08)*0.02 + 0.0
        # X = 1.04 + 0.0408 = 1.0808 <--- previous value 1.0808
        # f(X) = 2.1616  <--- return 2.1616

    def test_fitzHughNagumo_gilzenrat_figure_2(self):
        # Isolate the FitzHughNagumo mechanism for testing and recreate figure 2 from the gilzenrat paper

        initial_v = 0.2
        initial_w = 0.0

        F = IntegratorMechanism(
            name='IntegratorMech-FitzHughNagumoFunction',
            function=FitzHughNagumoIntegrator(
                initial_v=initial_v,
                initial_w=initial_w,
                time_step_size=0.01,
                time_constant_w=1.0,
                time_constant_v=0.01,
                a_v=-1.0,
                b_v=1.0,
                c_v=1.0,
                d_v=0.0,
                e_v=-1.0,
                f_v=1.0,
                threshold=0.5,
                mode=1.0,
                uncorrelated_activity=0.0,
                a_w=1.0,
                b_w=-1.0,
                c_w=0.0

            )
        )
        plot_v_list = [initial_v]
        plot_w_list = [initial_w]

        # found this stimulus by guess and check b/c one was not provided with Figure 2 params
        stimulus = 0.073
        # increase range to 200 to match Figure 2 in Gilzenrat
        for i in range(10):
            results = F.execute(stimulus)
            plot_v_list.append(results[0][0][0])
            plot_w_list.append(results[1][0][0])

        # ** uncomment the lines below if you want to view the plot:
        # from matplotlib import pyplot as plt
        # plt.plot(plot_v_list)
        # plt.plot(plot_w_list)
        # plt.show()

        np.testing.assert_allclose(plot_v_list, [0.2, 0.22493312915681499, 0.24840327807265583, 0.27101619694032797,
                                                 0.29325863380332173, 0.31556552465130933, 0.33836727470568129,
                                                 0.36212868305470697, 0.38738542852040492, 0.41478016676749552,
                                                 0.44509530539552955]
)
        print(plot_w_list)
        np.testing.assert_allclose(plot_w_list, [0.0, 0.0019900332500000003, 0.0042083541185625045,
                                                 0.0066381342093118408, 0.009268739886338381, 0.012094486544132229,
                                                 0.015114073825358726, 0.018330496914962583, 0.021751346023501487,
                                                 0.025389465931011893, 0.029263968140538919]

)
#
# class TestGilzenratFullModel:
#     def test_replicate_gilzenrat_paper(self):
#         """
#         This implements a model of Locus Coeruleus / Norepinephrine (LC/NE) function described in `Gilzenrat et al. (2002)
#         <http://www.sciencedirect.com/science/article/pii/S0893608002000552?via%3Dihub>`_, used to simulate behavioral
#         and electrophysiological data (from LC recordings) in non-human primates.
#
#         This test does NOT validate output against expected values yet -- it is only here to ensure that changes to
#         PsyNeuLink do not prevent the system below from being created without errors
#
#         Plotting code is commented out and only one trial will be executed
#         """
#         # --------------------------------- Global Variables ----------------------------------------
#
#         # Mode ("coherence")
#         C = 0.95
#         # Uncorrelated Activity
#         d = 0.5
#
#         # Initial values
#         initial_h_of_v = 0.07
#         # initial_h_of_v = 0.07
#         initial_v = (initial_h_of_v - (1 - C) * d) / C
#         # initial_w = 0.14
#         initial_w = 0.14
#
#         # g(t) = G + k*w(t)
#
#         # Scaling factor for transforming NE release (u ) to gain (g ) on potentiated units
#         k = 3.0
#         # Base level of gain applied to decision and response units
#         G = 0.5
#
#         # numerical integration
#         time_step_size = 0.02
#         # number_of_trials = int(20/time_step_size)
#         number_of_trials = 1
#
#         # noise
#         standard_deviation = 0.22 * (time_step_size ** 0.5)
#
#         # --------------------------------------------------------------------------------------------
#
#         input_layer = TransferMechanism(default_variable=np.array([[0, 0]]),
#                                         name='INPUT LAYER')
#
#         # Implement projections from inputs to decision layer with weak cross-talk connections
#         #    from target and distractor inputs to their competing decision layer units
#         input_weights = np.array([[1, .33], [.33, 1]])
#
#         # Implement self-excitatory (auto) and mutually inhibitory (hetero) connections within the decision layer
#         decision_layer = GilzenratTransferMechanism(input_shapes=2,
#                                                     initial_value=np.array([[1, 0]]),
#                                                     matrix=np.array([[1, 0], [0, -1]]),
#                                                     # auto=1.0,
#                                                     # hetero=-1.0,
#                                                     time_step_size=time_step_size,
#                                                     noise=NormalDist(mean=0.0,
#                                                                      standard_deviation=standard_deviation).function,
#                                                     function=Logistic(bias=0.0),
#                                                     name='DECISION LAYER')
#
#         # Implement connection from target but not distractor unit in decision layer to response
#         output_weights = np.array([[1.84], [0]])
#
#         # Implement response layer with a single, self-excitatory connection
#         # To do Markus: specify recurrent self-connrection weight for response unit to 2.00
#         response = GilzenratTransferMechanism(input_shapes=1,
#                                               initial_value=np.array([[2.0]]),
#                                               matrix=np.array([[0.5]]),
#                                               function=Logistic(bias=2),
#                                               time_step_size=time_step_size,
#                                               noise=NormalDist(mean=0.0, standard_deviation=standard_deviation).function,
#                                               name='RESPONSE')
#
#         # Implement response layer with input_port for ObjectiveMechanism that has a single value
#         # and a MappingProjection to it that zeros the contribution of the decision unit in the decision layer
#         LC = LCControlMechanism(
#             time_step_size_FitzHughNagumo=time_step_size,  # integrating step size
#             mode_FitzHughNagumo=C,  # coherence: set to either .95 or .55
#             uncorrelated_activity_FitzHughNagumo=d,  # Baseline level of intrinsic, uncorrelated LC activity
#             time_constant_v_FitzHughNagumo=0.05,
#             time_constant_w_FitzHughNagumo=5,
#             a_v_FitzHughNagumo=-1.0,
#             b_v_FitzHughNagumo=1.0,
#             c_v_FitzHughNagumo=1.0,
#             d_v_FitzHughNagumo=0.0,
#             e_v_FitzHughNagumo=-1.0,
#             f_v_FitzHughNagumo=1.0,
#             a_w_FitzHughNagumo=1.0,
#             b_w_FitzHughNagumo=-1.0,
#             c_w_FitzHughNagumo=0.0,
#             t_0_FitzHughNagumo=0,
#             initial_v_FitzHughNagumo=initial_v,
#             initial_w_FitzHughNagumo=initial_w,
#             threshold_FitzHughNagumo=0.5,
#             # Parameter describing shape of the FitzHugh–Nagumo cubic nullcline for the fast excitation variable v
#             objective_mechanism=ObjectiveMechanism(
#                 function=Linear,
#                 monitor=[(decision_layer, None, None, np.array([[0.3], [0.0]]))],
#                 # monitor=[{PROJECTION_TYPE: MappingProjection,
#                 #                           SENDER: decision_layer,
#                 #                           MATRIX: np.array([[0.3],[0.0]])}],
#                 name='LC ObjectiveMechanism'
#             ),
#             modulated_mechanisms=[decision_layer, response],
#             name='LC')
#
#         for signal in LC._control_signals:
#             signal._intensity = k * initial_w + G
#
#         # ELICITS WARNING:
#         decision_process = Process(pathway=[input_layer,
#                                             input_weights,
#                                             decision_layer,
#                                             output_weights,
#                                             response],
#                                    name='DECISION PROCESS')
#
#         lc_process = Process(pathway=[decision_layer,
#                                       # CAUSES ERROR:
#                                       # np.array([[1,0],[0,0]]),
#                                       LC],
#                              name='LC PROCESS')
#
#         task = System(processes=[decision_process, lc_process])
#
#         # stimulus
#         stim_list_dict = {input_layer: np.repeat(np.array([[0, 0], [1, 0]]), 10 / time_step_size, axis=0)}
#
#         def h_v(v, C, d):
#             return C * v + (1 - C) * d
#
#         # Initialize output arrays for plotting
#         LC_results_v = [h_v(initial_v, C, d)]
#         LC_results_w = [initial_w]
#         decision_layer_target = [0.5]
#         decision_layer_distractor = [0.5]
#         response_layer = [0.5]
#
#         def record_trial():
#             LC_results_v.append(h_v(LC.value[2][0], C, d))
#             LC_results_w.append(LC.value[3][0])
#             decision_layer_target.append(decision_layer.value[0][0])
#             decision_layer_distractor.append(decision_layer.value[0][1])
#             response_layer.append(response.value[0][0])
#             current_trial_num = len(LC_results_v)
#             if current_trial_num % 50 == 0:
#                 percent = int(round((float(current_trial_num) / number_of_trials) * 100))
#                 sys.stdout.write("\r" + str(percent) + "% complete")
#                 sys.stdout.flush()
#
#         sys.stdout.write("\r0% complete")
#         sys.stdout.flush()
#         task.run(stim_list_dict, num_trials=number_of_trials, call_after_trial=record_trial)
#
#         # from matplotlib import pyplot as plt
#         # import numpy as np
#         # t = np.arange(0.0, len(LC_results_v), 1.0)
#         # plt.plot(t, LC_results_v, label="h(v)")
#         # plt.plot(t, LC_results_w, label="w")
#         # plt.plot(t, decision_layer_target, label="target")
#         # plt.plot(t, decision_layer_distractor, label="distractor")
#         # plt.plot(t, response_layer, label="response")
#         # plt.xlabel(' # of timesteps ')
#         # plt.ylabel('h(V)')
#         # plt.legend(loc='upper left')
#         # plt.ylim((-0.2, 1.2))
#         # plt.show()
#
#         # This prints information about the System,
#         # including its execution list indicating the order in which the Mechanisms will execute
#         # IMPLEMENTATION NOTE:
#         #  MAY STILL NEED TO SCHEDULE RESPONSE TO EXECUTE BEFORE LC
#         #  (TO BE MODULATED BY THE GAIN MANIPULATION IN SYNCH WITH THE DECISION LAYER
#         # task.show()
#
#         # This displays a diagram of the System
#         # task.show_graph()
#



import numpy as np
import psyneulink.core.components.functions.function as Function
import psyneulink.core.components.functions.nonstateful.objectivefunctions as Functions
import psyneulink.core.components.functions.nonstateful.optimizationfunctions as OPTFunctions
import psyneulink.core.globals.keywords as kw
from psyneulink.core.globals.sampleiterator import SampleIterator, SampleSpec
import pytest

SIZE=5
# Some metrics (CROSS_ENTROPY) don't like 0s
test_var = np.random.rand(SIZE) + Function.EPSILON
EPS = float(Function.EPSILON)
search_space = [SampleIterator([EPS, 1.0] if i % 2 == 0 else SampleSpec(start=EPS, stop=1.0, num=2)) for i in range(SIZE)]
results = {
    Functions.Stability: {
        kw.ENERGY: {
            True: {
                OPTFunctions.MINIMIZE: {
                    'FIRST': ((1.0, 1.0, 1.0, 1.0, 1.0), -0.4, [], []),
                    'RANDOM': ((1.0, 1.0, 1.0, 1.0, 1.0), -0.4, [], []),
                },
                OPTFunctions.MAXIMIZE: {
                    'FIRST': ((EPS, EPS, EPS, EPS, EPS), -1.9721522630525296e-32, [], []),
                    'RANDOM': ((1.0, EPS, EPS, EPS, EPS), -1.9721522630525296e-32, [], []),
                },
            },
            False: {
                OPTFunctions.MINIMIZE: {
                    'FIRST': ((1.0, 1.0, 1.0, 1.0, 1.0), -10.0, [], []),
                    'RANDOM': ((1.0, 1.0, 1.0, 1.0, 1.0), -10.0, [], []),
                },
                OPTFunctions.MAXIMIZE: {
                    'FIRST': ((EPS, EPS, EPS, EPS, EPS), -4.930380657631324e-31, [], []),
                    'RANDOM': ((1.0, EPS, EPS, EPS, EPS), -4.930380657631324e-31, [], []),
                },
            },
        },
        kw.ENTROPY: {
            True: {
                OPTFunctions.MINIMIZE: {
                    'FIRST': ((1.0, 1.0, 1.0, 1.0, 1.0), -1.3862943611198906, [], []),
                    'RANDOM': ((1.0, 1.0, 1.0, 1.0, 1.0), -1.3862943611198906, [], []),
                },
                OPTFunctions.MAXIMIZE: {
                    'FIRST': ((EPS, EPS, EPS, EPS, 1.0), 6.931471805599453, [], []),
                    'RANDOM': ((EPS, EPS, 1.0, EPS, EPS), 6.931471805599453, [], []),
                },
            },
            False: {
                OPTFunctions.MINIMIZE: {
                    'FIRST': ((1.0, 1.0, 1.0, 1.0, 1.0), -6.931471805599453, [], []),
                    'RANDOM': ((1.0, 1.0, 1.0, 1.0, 1.0), -6.931471805599453, [], []),
                },
                OPTFunctions.MAXIMIZE: {
                    'FIRST': ((EPS, EPS, EPS, EPS, 1.0), 34.657359027997266, [], []),
                    'RANDOM': ((EPS, EPS, 1.0, EPS, EPS), 34.657359027997266, [], []),
                },
            },
        },
    },
}


@pytest.mark.function
@pytest.mark.benchmark
@pytest.mark.optimization_function
@pytest.mark.parametrize("selection", ['FIRST', 'RANDOM'])
@pytest.mark.parametrize("direction", [OPTFunctions.MINIMIZE, OPTFunctions.MAXIMIZE])
@pytest.mark.parametrize("normalize", [True, False])
@pytest.mark.parametrize("metric", [kw.ENERGY, kw.ENTROPY])
@pytest.mark.parametrize("obj_func", [Functions.Stability])
def test_grid_search(obj_func, metric, normalize, direction, selection, benchmark, func_mode):
    variable = test_var
    result = results[obj_func][metric][normalize][direction][selection]
    benchmark.group = "OptimizationFunction " + str(obj_func) + " " + metric

    of = obj_func(default_variable=variable, metric=metric, normalize=normalize)
    f = OPTFunctions.GridSearch(objective_function=of, default_variable=variable,
                                search_space=search_space, direction=direction,
                                select_randomly_from_optimal_values=(selection=='RANDOM'),
                                seed=0, save_values=False)
    EX = pytest.helpers.get_func_execution(f, func_mode)

    res = benchmark(EX, variable)

    np.testing.assert_allclose(res[0], result[0], rtol=1e-5, atol=1e-8)
    np.testing.assert_allclose(res[1], result[1], rtol=1e-5, atol=1e-8)
    if func_mode == 'Python':
        np.testing.assert_allclose(res[2], result[2], rtol=1e-5, atol=1e-8)
        np.testing.assert_allclose(res[3], result[3], rtol=1e-5, atol=1e-8)



from .regressioncfa import *
from .compositionrunner import *
from .autodiffcomposition import *
from .emcomposition import *
__all__ = list(regressioncfa.__all__)
__all__.extend(compositionrunner.__all__)
__all__.extend(autodiffcomposition.__all__)
__all__.extend(emcomposition.__all__)

try:
    import torch
    from torch import nn

    # Some torch releases have silent dependency on a more recent numpy than the one curently required by PNL.
    # This breaks torch numpy bindings, see e.g:  https://github.com/pytorch/pytorch/issues/100690
    torch.tensor([1,2,3]).numpy()

    torch_available = True
except (ImportError, RuntimeError):
    torch_available = False

__all__.append('torch_available')



import numpy as np
import pytest

from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism

VECTOR_SIZE=4

class TestObjectiveMechanism:
    # VALID INPUTS

    @pytest.mark.mechanism
    @pytest.mark.objective_mechanism
    @pytest.mark.benchmark(group="ObjectiveMechanism")
    def test_objective_mech_inputs_list_of_ints(self, benchmark, mech_mode):

        O = ObjectiveMechanism(
            name='O',
            default_variable=[0 for i in range(VECTOR_SIZE)],
        )
        EX = pytest.helpers.get_mech_execution(O, mech_mode)

        val = benchmark(EX, [10.0 for i in range(VECTOR_SIZE)])
        np.testing.assert_allclose(val, [[10.0 for i in range(VECTOR_SIZE)]])



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *******************************************  TIMER FUNCTIONS  *****************************************************
"""

* `TimerFunction`
* `LinearTimer`
* `AcceleratingTimer`
* `DeceleratingTimer`
* `AsymptoticTimer`

Overview
--------

Functions for which `initial <TimerFunction.initial>` and `final <TimerFunction.final>` values and a `duration
<TimerFunction.duration>` can be specified, for use with a `TimerMechanism`.

.. _TimerFunction_Types:

Types
~~~~~

There are four types that implement different functional forms, each of which is rising if `initial
<TimerFunction.initial>` is less than `final <TimerFunction.final>` and declining for the reverse:

* **LinearTimer** - progresses linearly from `initial <TimerFunction.initial>` to `final <TimerFunction.final>` value.
  (see `interactive graph <https://www.desmos.com/calculator/i0knnnozcs>`_).

* **AcceleratingTimer** - advances from initial <TimerFunction.initial>` to `final <TimerFunction.final>` value
  by progressively larger amounts at an adjustable exponential `rate <AcceleratingTimerRise.rate>`
  (see `interactive graph <https://www.desmos.com/calculator/rms6z2ji8g>`_).

* **DeceleratingTimer** - advances from initial <TimerFunction.initial>` to `final <TimerFunction.final>` value
  by progressively smaller amounts at an adjustable exponential `rate <DeceleratingTimer.rate>`
  (see `interactive graph <https://www.desmos.com/calculator/cshkzip0ai>`_).

* **AsymptoticTimer** - progresses at a fixed exponential `rate <AsymptoticTimer.rate>` from `initial
  <TimerFunction.initial>` to within `tolerance <AsymptoticTimer.tolerance>` of `final <TimerFunction.final>`
    (see `interactive graph <https://www.desmos.com/calculator/tmfs4ps9cp>`_).


.. _TimerFunction_StandardAttributes:

Standard Attributes
~~~~~~~~~~~~~~~~~~~

TimerFunctions have the following Parameters:

.. _TimerFunction_Initial:
* **initial**: specifies the `value <Function_Base.value>` that the function has when its `variable
  <Function_Base.variable>` is 0.

.. _TimerFunction_Final:
* **final**: specifies the `value <Function_Base.value>` that the function has when its `variable
  <Function_Base.variable>` is equal to `duration <TimerFunction.duration>`.

.. _TimerFunction_Duration:
* **duration**: specifies the value of the `variable <Function_Base.variable>` at which the`value
  <Function_Base.value>` of the function is equal to `final <TimerFunction.final>`.

.. _TimerFunction_Rate:
* **rate**: specifies the rate at which the progression of the `value <Function_Base.value>` of the function changes.

TimerFunction Class References
------------------------------

"""

from math import e

import numpy as np
try:
    import torch
except ImportError:
    torch = None
from beartype import beartype

from psyneulink._typing import Optional

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.functions.nonstateful.transferfunctions import TransferFunction
from psyneulink.core.globals.context import handle_external_context
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.utilities import (ValidParamSpecType)
from psyneulink.core.globals.preferences.basepreferenceset import \
    REPORT_OUTPUT_PREF, PreferenceEntry, PreferenceLevel, ValidPrefSet
from psyneulink.core.globals.keywords import \
    (ADDITIVE_PARAM, ACCELERATING_TIMER_FUNCTION, ASYMPTOTIC_TIMER_FUNCTION, DECELERATING_TIMER_FUNCTION,
     DURATION, FINAL, INITIAL, LINEAR_TIMER_FUNCTION, MULTIPLICATIVE_PARAM, OFFSET, PREFERENCE_SET_NAME,
     RATE, TIMER_FUNCTION, TIMER_FUNCTION_TYPE, TOLERANCE)

__all__ = ['LinearTimer','AcceleratingTimer','DeceleratingTimer','AsymptoticTimer']


class TimerFunction(TransferFunction):  # --------------------------------------------------------------------------------
    """Subclass of TransferFunction that allows a initial, final and duration value to be specified;
    for use with a `TimerMechanism`.

    Attributes
    ----------

    variable : number or array
        contains value to be transformed.

    initial : float (>0)
        determines the value of the function when `variable <TimerFunction.variable>` = 0.

    final : float
        determines the value of the function when `variable <TimerFunction.variable>` = `duration <TimerFunction.duration>`.

    duration : float (>0)
        determines the value of `variable <TimerFunction.variable>` at which the value of the function is equal
        to `final <TimerFunction.final>`.

    rate : float (>1.0)
        determines the rate at which the value of the function accelerates.

    """
    componentType = TIMER_FUNCTION_TYPE
    componentName = TIMER_FUNCTION

    class Parameters(TransferFunction.Parameters):
        """
            Attributes
            ----------

                duration
                    see `duration <TimerFunction.initial>`

                    :default value: None
                    :type: 'float'

                final
                    see `final <TimerFunction.final>`

                    :default value: None
                    :type: 'float'

                initial
                    see `initial <TimerFunction.initial>`

                    :default value: None
                    :type: 'float'

                rate
                    see `rate <TimerFunction.rate>`

                    :default value: None
                    :type: 'float'
        """
        initial = Parameter(1.0, modulable=True)
        final = Parameter(0.0, modulable=True)
        duration = Parameter(1.0, modulable=True, aliases=[MULTIPLICATIVE_PARAM])
        rate = Parameter(1.0, modulable=True)

        def _validate_duration(self, duration):
            if duration <= 0:
                return f"must be greater than 0."

        def _validate_rate(self, rate):
            if rate < 1:
                return f"must be greater than or equal to 1.0."


class LinearTimer(TimerFunction):
    """LinearTimer(        \
         default_variable, \
         initial=0.0,      \
         final=1.0,        \
         duration=1.0,     \
         params=None,      \
         owner=None,       \
         name=None,        \
         prefs=None        \
         )

    .. _LinearTimer:
    |
    `function <LinearTimer._function>` returns linear transform of `variable <LinearTimer.variable>`.

    .. math::
       \\left(\\frac{final-initial}{duration}\\right) \\cdot variable + initial

    such that:

    .. math::
        value=initial \\ for\\ variable=0

        value=final\\ for\\ variable=duration

    where:

        **initial** determines the `value <Function_Base.value>` of the function
        when its `variable <LinearTimer.variable>` = 0.

        **final** determines the `value <Function_Base.value>` of the function
        when its `variable <LinearTimer.variable>` = duration.

        **duration** determines the value of `variable <LinearTimer.variable>`
        at which the value of the function = final.

    `derivative <LinearTimer.derivative>` returns the derivative of the LinearTimer Function:

      .. math::
         \\frac{final-initial}{duration}

    See `graph <https://www.desmos.com/calculator/i0knnnozcs>`_ for interactive plot of the function using `Desmos
    <https://www.desmos.com>`_.

    Arguments
    ---------

    default_variable : number or array : default class_defaults.variable
        specifies a template for the value to be transformed.

    initial : float : default 1.0
        specifies the value the function should have when `variable <LinearTimer.variable>` = 0;
        must be greater than or equal to 0.

    final : float : default 1.0
        specifies the value the function should have when `variable <LinearTimer.variable>` = `duration
        <TimerFunction.duration>`; must be greater than `initial <LinearTimer.initial>`.

    duration : float : default 1.0
        specifies the value of `variable <LinearTimer.variable>` at which the value of the function
        should equal `final <LinearTimer.final>`; must be greater than 0.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters
        in arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or array
        contains value to be transformed.

    initial : float (>0)
        determines the value of the function when `variable <LinearTimer.variable>` = 0.

    final : float
        determines the value of the function when `variable <LinearTimer.variable>` = `duration <TimerFunction.duration>`.

    duration : float (>0)
        determines the value of `variable <LinearTimer.variable>` at which the value of the function is equal
        to `final <LinearTimer.final>`.

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = LINEAR_TIMER_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'LinearTimerClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    _model_spec_class_name_is_generic = True

    # FIX: REINSTATE Parameters AND VALIDATE INITIAL, FINAL AND DURATION

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 initial: Optional[ValidParamSpecType] = None,
                 final: Optional[ValidParamSpecType] = None,
                 duration: Optional[ValidParamSpecType] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):
        super().__init__(
            default_variable=default_variable,
            initial=initial,
            final=final,
            duration=duration,
            params=params,
            owner=owner,
            prefs=prefs,
        )

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """

        Arguments
        ---------

        variable : number or array : default class_defaults.variable
           a single value or array to be exponentiated.

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.

        Returns
        -------

        LinearTimer rise transform of variable : number or array

        """
        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        result = ((final - initial) / duration) * variable + initial
        return self.convert_output_type(result)

    @handle_external_context()
    def derivative(self, input, output=None, context=None):
        """Derivative of `function <LinearTimer._function>` at **input**:

        .. math::
           (final - initial) * \\left(\\frac{(1 + duration * e^{variable})}{duration * e^{duration}}\\right)

        Arguments
        ---------

        input : number
            value of the input to the LinearTimer transform at which derivative is to be taken.

        Returns
        -------
        derivative :  number or array
        """
        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        return (final - initial) / duration

    # FIX:
    def _gen_llvm_transfer(self, builder, index, ctx, vi, vo, params, state, *, tags:frozenset):
        ptri = builder.gep(vi, [ctx.int32_ty(0), index])
        ptro = builder.gep(vo, [ctx.int32_ty(0), index])

        duration_ptr = ctx.get_param_or_state_ptr(builder, self, DURATION, param_struct_ptr=params)
        initial_ptr = ctx.get_param_or_state_ptr(builder, self, INITIAL, param_struct_ptr=params)
        final_ptr = ctx.get_param_or_state_ptr(builder, self, FINAL, param_struct_ptr=params)
        offset_ptr = ctx.get_param_or_state_ptr(builder, self, OFFSET, param_struct_ptr=params)

        initial = pnlvm.helpers.load_extract_scalar_array_one(builder, initial_ptr)
        final = pnlvm.helpers.load_extract_scalar_array_one(builder, final_ptr)
        duration = pnlvm.helpers.load_extract_scalar_array_one(builder, duration_ptr)
        offset = pnlvm.helpers.load_extract_scalar_array_one(builder, offset_ptr)

        exp_f = ctx.get_builtin("exp", [ctx.float_ty])
        val = builder.load(ptri)
        val = builder.fmul(val, duration)
        val = builder.fadd(val, initial)
        val = builder.call(exp_f, [val])

        if "derivative" in tags:
            # f'(x) = s*r*e^(r*x + b)
            val = builder.fmul(val, final)
            val = builder.fmul(val, duration)
        else:
            # f(x) = s*e^(r*x + b) + o
            val = builder.fmul(val, final)
            val = builder.fadd(val, offset)

        builder.store(val, ptro)

    # FIX:
    def _gen_pytorch_fct(self, device, context=None):
        final = self._get_pytorch_fct_param_value(FINAL, device, context)
        initial = self._get_pytorch_fct_param_value(INITIAL, device, context)
        duration = self._get_pytorch_fct_param_value(DURATION, device, context)
        return lambda x : ((final - initial) / duration) * x + initial


class AcceleratingTimer(TimerFunction):
    """
    AcceleratingTimer(     \
         default_variable, \
         initial=0.0,      \
         final=1.0,        \
         duration=1.0,     \
         rate=1.0,         \
         params=None,      \
         owner=None,       \
         name=None,        \
         prefs=None        \
         )

    .. _AcceleratingTimer:
    |
    `function <AcceleratingTimer._function>` returns acceleratingTimer rise transform of `variable
    <AcceleratingTimer.variable>`; this is the inverse of the `AcceleratingTimer` Function.

    .. math::
       initial+\\left(final-initial\\right)\\left(\\frac{variable}{duration}\\right)^{rate}e^{\\left(\\left(
       \\frac{variable}{duration}\\right)^{rate}-1\\right)}

    such that:

    .. math::
        value=initial \\ for\\ variable=0

        value=final\\ for\\ variable=duration

    where:

        **initial** determines the `value <Function_Base.value>` of the function
        when its `variable <AcceleratingTimer.variable>` = 0.

        **final** determines the `value <Function_Base.value>` of the function
        when its `variable <AcceleratingTimer.variable>` = duration.

        **duration** determines the value of `variable <AcceleratingTimer.variable>`
        at which the value of the function = final.

        **rate** determines the `rate <AcceleratingTimer.rate>` of acceleration of the function.

    `derivative <AcceleratingTimer.derivative>` returns the derivative of the AcceleratingTimer Function:

      .. math::
         (final-initial) \\cdot \\left[ rate \\cdot \\left(\\frac{variable}{duration}\\right)^{rate-1}
         \\cdot \\frac{1}{duration} \\cdot e^{\\left(\\left(\\frac{variable}{duration}\\right)^{rate}-1\\right)} +
         \\left(\\frac{variable}{duration}\\right)^{rate} \\cdot e^{\\left(\\left(\\frac{variable}{duration}\\right)^{
         rate}-1\\right)} \\cdot rate \\cdot \\frac{1}{duration}\\right]

    See `graph <https://www.desmos.com/calculator/rms6z2ji8g>`_ for interactive plot of the function using `Desmos
    <https://www.desmos.com>`_.

    Arguments
    ---------

    default_variable : number or array : default class_defaults.variable
        specifies a template for the value to be transformed.

    initial : float : default 1.0
        specifies the value the function should have when `variable <AcceleratingTimer.variable>` = 0;
        must be greater than or equal to 0.

    final : float : default 1.0
        specifies the value the function should have when `variable <AcceleratingTimer.variable>` = `duration
        <TimerFunction.duration>`; must be greater than `initial <AcceleratingTimer.initial>`.

    duration : float : default 1.0
        specifies the value of `variable <AcceleratingTimer.variable>` at which the value of the function
        should equal `final <AcceleratingTimer.final>`; must be greater than 0.

    rate : float : default 1.0
        specifies the rate at which the value of the function accelerates; must be greater than 0.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters
        in arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or array
        contains value to be transformed.

    initial : float (>0)
        determines the value of the function when `variable <AcceleratingTimer.variable>` = 0.

    final : float
        determines the value of the function when `variable <AcceleratingTimer.variable>` = `duration <TimerFunction.duration>`.

    duration : float (>0)
        determines the value of `variable <AcceleratingTimer.variable>` at which the value of the function is equal
        to `final <AcceleratingTimer.final>`.

    rate : float (>1.0)
        determines the rate at which the value of the function accelerates.

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = ACCELERATING_TIMER_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'AcceleratingTimerClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    _model_spec_class_name_is_generic = True

    # FIX: REINSTATE Parameters AND VALIDATE INITIAL, FINAL AND DURATION

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 initial: Optional[ValidParamSpecType] = None,
                 final: Optional[ValidParamSpecType] = None,
                 duration: Optional[ValidParamSpecType] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):
        super().__init__(
            default_variable=default_variable,
            initial=initial,
            final=final,
            duration=duration,
            params=params,
            owner=owner,
            prefs=prefs,
        )

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """

        Arguments
        ---------

        variable : number or array : default class_defaults.variable
           a single value or array to be exponentiated.

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.

        Returns
        -------

        AcceleratingTimer rise transform of variable : number or array

        """
        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        rate = self._get_current_parameter_value(RATE, context)

        result = (initial + (final - initial) * np.power((variable / duration),rate)
                  * np.exp(np.power((variable / duration),rate) - 1))

        return self.convert_output_type(result)

    @handle_external_context()
    def derivative(self, input, output=None, context=None):
        """Derivative of `function <AcceleratingTimer._function>` at **input**:

        .. math::
           (final - initial) * \\left(\\frac{(1 + duration * e^{variable})}{duration * e^{duration}}\\right)

        Arguments
        ---------

        input : number
            value of the input to the AcceleratingTimer transform at which derivative is to be taken.

        Returns
        -------
        derivative :  number or array
        """
        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        rate = self._get_current_parameter_value(RATE, context)

        return ((final - initial) *
                (rate * np.power((input / duration), (rate - 1))
                 * ((1 / duration)
                    * (np.exp(np.power((input / duration), rate) - 1)
                       + (np.power((input / duration), rate))
                       * np.exp(np.power((input / duration), rate) - 1) * rate * 1 / duration))))

    # FIX:
    def _gen_llvm_transfer(self, builder, index, ctx, vi, vo, params, state, *, tags:frozenset):
        ptri = builder.gep(vi, [ctx.int32_ty(0), index])
        ptro = builder.gep(vo, [ctx.int32_ty(0), index])

        duration_ptr = ctx.get_param_or_state_ptr(builder, self, DURATION, param_struct_ptr=params)
        initial_ptr = ctx.get_param_or_state_ptr(builder, self, INITIAL, param_struct_ptr=params)
        final_ptr = ctx.get_param_or_state_ptr(builder, self, FINAL, param_struct_ptr=params)
        offset_ptr = ctx.get_param_or_state_ptr(builder, self, OFFSET, param_struct_ptr=params)

        duration = pnlvm.helpers.load_extract_scalar_array_one(builder, duration_ptr)
        initial = pnlvm.helpers.load_extract_scalar_array_one(builder, initial_ptr)
        final = pnlvm.helpers.load_extract_scalar_array_one(builder, final_ptr)
        offset = pnlvm.helpers.load_extract_scalar_array_one(builder, offset_ptr)

        exp_f = ctx.get_builtin("exp", [ctx.float_ty])
        val = builder.load(ptri)
        val = builder.fmul(val, duration)
        val = builder.fadd(val, initial)
        val = builder.call(exp_f, [val])

        if "derivative" in tags:
            # f'(x) = s*r*e^(r*x + b)
            val = builder.fmul(val, final)
            val = builder.fmul(val, duration)
        else:
            # f(x) = s*e^(r*x + b) + o
            val = builder.fmul(val, final)
            val = builder.fadd(val, offset)

        builder.store(val, ptro)

    # FIX:
    def _gen_pytorch_fct(self, device, context=None):
        final = self._get_pytorch_fct_param_value(FINAL, device, context)
        initial = self._get_pytorch_fct_param_value(INITIAL, device, context)
        duration = self._get_pytorch_fct_param_value(DURATION, device, context)
        rate = self._get_pytorch_fct_param_value(RATE, device, context)

        return lambda x : (initial + (final - initial) * torch.power((x / duration),rate)
                           * torch.exp(torch.power((x / duration),rate) - 1))


class DeceleratingTimer(TimerFunction):  # ---------------------------------------------------------------------------
    """
    DeceleratingTimer(     \
         default_variable, \
         initial=1.0,      \
         duration=1.0,     \
         final=0.01,       \
         rate=1.0,         \
         params=None,      \
         owner=None,       \
         name=None,        \
         prefs=None        \
         )

    .. _DeceleratingTimer:
    |
    `function <DeceleratingTimer._function>` returns exponentially decaying transform of `variable
    <DeceleratingTimer.variable>`:

    .. math::
       \\frac{\\left(initial-final-direction\\right)}{e^{\\ln\\left(-direction\\left(initial\\ -\\
       final-direction\\right)\\right)\\left(\\frac{variable}{duration}\\right)^{
       rate}}}+final+direction

    such that:

    .. math::
        value = initial + offset\\ for\\ variable=0

        value = (initial * final) + offset\\ for\\ variable=duration

    where:

        **initial**, together with `offset <DeceleratingTimer.offset>`, determines the value of the function when
        `variable <DeceleratingTimer.variable>` = 0, and is used together with `final <DeceleratingTimer.final>`
        to determine the value of the function when `variable <DeceleratingTimer.variable>` = `duration
        <DeceleratingTimer.duration>`.

        **duration** determines the value of `variable <DeceleratingTimer.variable>` at which
        the value of the function should equal :math:`initial * final + offset`.

        **final** is the fraction of `initial <DeceleratingTimer.initial>` when, added to `offset
        <DeceleratingTimer.offset>`, is used to determine the value of the function when `variable
        <DeceleratingTimer.variable>` should equal `duration <DeceleratingTimer.duration>`.

        **rate** determines the `rate <DeceleratingTimer.rate>` of deceleration of the function.

        **direction** is +1 if final > initial, otherwise -1, and is used to determine the direction of the
        progression (rising or decaying) of the TimerFunction.

    `derivative <DeceleratingTimer.derivative>` returns the derivative of the DeceleratingTimer Function:

      .. math::
         \\frac{direction \\cdot rate \\cdot(initial-final-direction)\\cdot\\ln(direction(final-initial+direction)) \\cdot \\left(\\frac{
         variable}{duration}\\right)^{rate-1}}{duration\\cdot e^{\\ln(direction(final-initial+direction))\\left(\\frac{variable}{
         duration}\\right)^{rate}}}

    See `graph <https://www.desmos.com/calculator/cshkzip0ai>`_ for interactive plot of the function using `Desmos
    <https://www.desmos.com>`_.


    Arguments
    ---------

    default_variable : number or array : default class_defaults.variable
        specifies a template for the value to be transformed.

    initial : float : default 1.0
        specifies, together with `offset <DeceleratingTimer.offset>`, the value of the function when `variable
        <DeceleratingTimer.variable>` = 0; must be greater than 0.

    final : float : default 0.01
        specifies the fraction of `initial <DeceleratingTimer.initial>` when added to `offset <DeceleratingTimer.offset>`,
        that determines the value of the function when `variable <DeceleratingTimer.variable>` = `duration
        <DeceleratingTimer.duration>`; must be between 0 and 1.

    duration : float : default 1.0
        specifies the value of `variable <DeceleratingTimer.variable>` at which the `value of the function
        should equal `initial <DeceleratingTimer.initial>` * `final <DeceleratingTimer.final>` + `offset
        <DeceleratingTimer.offset>`; must be greater than 0.

    rate : float : default 1.0
        specifies the rate at which the value of the function decelerates; must be greater than 0.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters
        in arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or array
        contains value to be transformed.

    initial : float
        determines, together with `offset <DeceleratingTimer.offset>`, the value of the function when `variable
        <DeceleratingTimer.variable>` = 0.

    final : float
        determines the fraction of `initial <DeceleratingTimer.initial>` when added to `offset <DeceleratingTimer.offset>`,
        that determines the value of the function when `variable <DeceleratingTimer.variable>` = `duration
        <DeceleratingTimer.duration>`.

    duration : float (>0)
        determines the value of `variable <DeceleratingTimer.variable>` at which the value of the function should
        equal `initial <DeceleratingTimer.initial>` * `final <DeceleratingTimer.final>` + `offset <DeceleratingTimer.offset>`.

    rate : float (>1.0)
        determines the rate at which the value of the function decelerates.

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = DECELERATING_TIMER_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'DeceleratingTimerClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    _model_spec_class_name_is_generic = True

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 initial: Optional[ValidParamSpecType] = None,
                 final: Optional[ValidParamSpecType] = None,
                 duration: Optional[ValidParamSpecType] = None,
                 rate: Optional[ValidParamSpecType] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):
        super().__init__(
            default_variable=default_variable,
            initial=initial,
            final=final,
            duration=duration,
            rate=rate,
            params=params,
            owner=owner,
            prefs=prefs,
        )

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """

        Arguments
        ---------

        variable : number or array : default class_defaults.variable
           amount by which to increment timer on current execution;  if this is not specified, the timer is incremented
           by the value of `increment <DeceleratingTimer.increment>`.

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.

        Returns
        -------

        Exponentially decayed value of variable : number or array

        """
        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        rate = self._get_current_parameter_value(RATE, context)

        direction = 1 if final > initial else -1

        result = ((initial - final - direction) /
                  np.exp((np.log(-direction * (initial - final - direction)) * np.power((variable / duration),rate)))
                  + final + direction)

        return self.convert_output_type(result)

    @handle_external_context()
    def derivative(self, input, output=None, context=None):
        """Derivative of `function <DeceleratingTimer._function>` at **input**:

      .. math::
         \\frac{direction \\cdot rate \\cdot(initial-final-direction)\\cdot\\ln(direction(
         final-initial+direction))\\cdot \\left(\\frac{variable}{duration}\\right)^{rate-1}}{duration\\cdot e^{\\ln(
         direction(final-initial+direction))\\left(\\frac{variable}{duration}\\right)^{rate}}}


        Arguments
        ---------

        input : number
            value of the input to the DeceleratingTimer transform at which derivative is to be taken.

        Derivative of `function <DeceleratingTimer._function>` at **input**.

        Returns
        -------
        derivative :  number or array
        """

        initial = self._get_current_parameter_value(INITIAL, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        rate = self._get_current_parameter_value(RATE, context)
        direction = 1 if final > initial else -1

        return (direction * rate * (initial - final - direction) * np.log(direction * (final - initial + direction)) *
                np.power((input / duration), (rate - 1))
                / (duration * np.exp(np.log(direction * (final - initial + direction)) *
                                     np.power((input / duration), rate))))

    # FIX:
    def _gen_llvm_transfer(self, builder, index, ctx, vi, vo, params, state, *, tags:frozenset):
        ptri = builder.gep(vi, [ctx.int32_ty(0), index])
        ptro = builder.gep(vo, [ctx.int32_ty(0), index])

        duration_ptr = ctx.get_param_or_state_ptr(builder, self, DURATION, param_struct_ptr=params)
        initial_ptr = ctx.get_param_or_state_ptr(builder, self, INITIAL, param_struct_ptr=params)
        final_ptr = ctx.get_param_or_state_ptr(builder, self, FINAL, param_struct_ptr=params)
        offset_ptr = ctx.get_param_or_state_ptr(builder, self, OFFSET, param_struct_ptr=params)

        duration = pnlvm.helpers.load_extract_scalar_array_one(builder, duration_ptr)
        initial = pnlvm.helpers.load_extract_scalar_array_one(builder, initial_ptr)
        final = pnlvm.helpers.load_extract_scalar_array_one(builder, final_ptr)
        offset = pnlvm.helpers.load_extract_scalar_array_one(builder, offset_ptr)

        exp_f = ctx.get_builtin("exp", [ctx.float_ty])
        val = builder.load(ptri)
        val = builder.fmul(val, duration)
        val = builder.fadd(val, initial)
        val = builder.call(exp_f, [val])

        if "derivative" in tags:
            # f'(x) = s*r*e^(r*x + b)
            val = builder.fmul(val, final)
            val = builder.fmul(val, duration)
        else:
            # f(x) = s*e^(r*x + b) + o
            val = builder.fmul(val, final)
            val = builder.fadd(val, offset)

        builder.store(val, ptro)

    def _gen_pytorch_fct(self, device, context=None):
        final = self._get_pytorch_fct_param_value(FINAL, device, context)
        initial = self._get_pytorch_fct_param_value(INITIAL, device, context)
        duration = self._get_pytorch_fct_param_value(DURATION, device, context)
        rate = self._get_pytorch_fct_param_value(RATE, device, context)
        direction = 1 if final > initial else -1

        return lambda x : ((initial - final - direction) /
                           (torch.log(-direction(initial - final - direction)) * torch.power((x / duration),rate))
                           + final + direction)


class AsymptoticTimer(TimerFunction):  # ---------------------------------------------------------------------------
    """
    AsymptoticTimer(       \
         default_variable, \
         initial=1.0,      \
         final=0,          \
         duration=1.0,     \
         tolerance=0.01,   \
         params=None,      \
         owner=None,       \
         name=None,        \
         prefs=None        \
         )

    .. AsymptoticTimer:
    |
    `function <AsymptoticTimer._function>` returns exponentially progressing transform of `variable
    <AsymptoticTimer.variable>` toward an asymptoticTimer value that reaches `duration <AsymptoticTimer.duration>`
    when it falls within the specified `tolerance <AsymptoticTimer.tolerance>` of `final
    <AsymptoticTimer.final>`:

    .. math::
       (initial - final) * \\frac{\\ln(tolerance)}{duration} *e^{\\left(\\frac{variable * \\ln(tolerance)}
       {duration}\\right)} + final

    such that:

    .. math::
        value = initial for\\ variable=0

        value = ((initial - final) \\cdot tolerance) for\\ variable=duration

    where:

        **initial**, determines the value of the function when `variable <AsymptoticTimer.variable>` = 0,
        and is used together with `final <AsymptoticTimer.final>` and `tolerance
        <AsymptoticTimer.tolerance>` to determine the value of the function at which `variable
        <AsymptoticTimer.variable>` = `duration <AsymptoticTimer.duration>`.

        **final** is the asymptoticTimer value toward which the function decays.

        **tolerance** is the fraction of `initial <AsymptoticTimer.initial>` - `final
        <AsymptoticTimer.final>` used to determine the value of the function when `variable
        <AsymptoticTimer.variable>` is equal to `duration <AsymptoticTimer.duration>`.

        **duration** determines the value of `variable <AsymptoticTimer.variable>` at which
        the value of the function is equal to :math:`initial \\cdot final`.

    .. _note::
       The function rises if `final <AsymptoticTimer.final>` > `initial <AsymptoticTimer.initial>` >,
       and decays if `final <AsymptoticTimer.final>` < `initial <AsymptoticTimer.initial>`.

    `derivative <AsymptoticTimer.derivative>` returns the derivative of the AsymptoticTimer Function:

      .. math::
         \\frac{initial\\cdot\\ln(tolerance)}{duration}\\cdot e^{\\frac{variable\\cdot\\ln(tolerance)}{duration}}

    See `graph <https://www.desmos.com/calculator/tmfs4ps9cp>`_ for interactive plot of the function using `Desmos
    <https://www.desmos.com>`_.

    Arguments
    ---------

    default_variable : number or array : default class_defaults.variable
        specifies a template for the value to be transformed.

    initial : float : default 1.0
        specifies the value of the function when `variable<AsymptoticTimer.variable>`=0; must be greater than 0.

    final : float : default 0.0
        specifies the asymptoticTimer value toward which the function decays.

    tolerance : float : default 0.01
        specifies the fraction of `initial <AsymptoticTimer.initial>`-`final <AsymptoticTimer.final>`
        that determines the value of the function when `variable <AsymptoticTimer.variable>` = `duration; must be
        between 0 and 1.

    duration : float : default 1.0
        specifies the value of `variable <AsymptoticTimer.variable>` at which the `value of the function
        should equal `initial <AsymptoticTimer.initial>` * `final <AsymptoticTimer.final>`;
        must be greater than 0.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters
        in arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or array
        contains value to be transformed.

    initial : float (>0)
        determines the value of the function when `variable <AsymptoticTimer.variable>` = 0.

    final : float
        determines the asymptoticTimer value toward which the function decays.

    tolerance : float (0,1)
        determines the fraction of `initial <AsymptoticTimer.initial>` - final <AsymptoticTimer.final>
        that determines the value of the function when `variable <AsymptoticTimer.variable>` = `duration
        <AsymptoticTimer.duration>`.

    duration : float (>0)
        determines the value of `variable <AsymptoticTimer.variable>` at which the value of the function should
        equal `initial <AsymptoticTimer.initial>` * `final <AsymptoticTimer.final>`.

    bounds : (None, None)

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = ASYMPTOTIC_TIMER_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'AsymptoticTimerClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    _model_spec_class_name_is_generic = True

    class Parameters(TimerFunction.Parameters):
        """
            Attributes
            ----------
                rate
                    see `tolerance <AsymptoticTimer.tolerance>`

                    :default value: None
                    :type: ``float``

                tolerance
                    see `tolerance <AsymptoticTimer.tolerance>`

                    :default value: 0.01
                    :type: ``float``
        """
        rate = Parameter(None)
        tolerance = Parameter(0.01, modulable=True)

        def _validate_rate(self, rate):
            if rate is not None:
                return f"is not used and should be left as None."

        def _validate_tolerance(self, tolerance):
            if tolerance <= 0 or tolerance >= 1:
                return f"must be between 0 and 1."

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 initial: Optional[ValidParamSpecType] = None,
                 final: Optional[ValidParamSpecType] = None,
                 tolerance: Optional[ValidParamSpecType] = None,
                 duration: Optional[ValidParamSpecType] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):
        super().__init__(
            default_variable=default_variable,
            initial=initial,
            duration=duration,
            final=final,
            tolerance=tolerance,
            params=params,
            owner=owner,
            prefs=prefs,
        )

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """

        Arguments
        ---------

        variable : number or array : default class_defaults.variable
           amount by which to increment timer on current execution;  if this is not specified, the timer is incremented
           by the value of `increment <AsymptoticTimer.increment>`.

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.

        Returns
        -------

        Exponentially decayed value of variable : number or array

        """
        initial = self._get_current_parameter_value(INITIAL, context)
        duration = self._get_current_parameter_value(DURATION, context)
        tolerance = self._get_current_parameter_value(TOLERANCE, context)
        final = self._get_current_parameter_value(FINAL, context)

        result = (initial - final) * np.exp(variable * np.log(tolerance) / duration) + final

        return self.convert_output_type(result)

    @handle_external_context()
    def derivative(self, input, output=None, context=None):
        """Derivative of `function <AsymptoticTimer._function>` at **input**:

        .. math::
           \\frac{initial\\cdot\\ln(tolerance)}{duration}\\cdot e^{\\frac{variable\\cdot\\ln(tolerance)}{duration}}

        Arguments
        ---------

        input : number
            value of the input to the AsymptoticTimer transform at which derivative is to be taken.

        Derivative of `function <AsymptoticTimer._function>` at **input**.

        Returns
        -------
        derivative :  number or array
        """

        initial = self._get_current_parameter_value(INITIAL, context)
        tolerance = self._get_current_parameter_value(TOLERANCE, context)
        final = self._get_current_parameter_value(FINAL, context)
        duration = self._get_current_parameter_value(DURATION, context)

        return (initial * np.log(tolerance) / duration) * np.exp(input * np.log(tolerance) / duration)

    # FIX:
    def _gen_llvm_transfer(self, builder, index, ctx, vi, vo, params, state, *, tags:frozenset):
        ptri = builder.gep(vi, [ctx.int32_ty(0), index])
        ptro = builder.gep(vo, [ctx.int32_ty(0), index])

        duration_ptr = ctx.get_param_or_state_ptr(builder, self, DURATION, param_struct_ptr=params)
        initial_ptr = ctx.get_param_or_state_ptr(builder, self, INITIAL, param_struct_ptr=params)
        final_ptr = ctx.get_param_or_state_ptr(builder, self, FINAL, param_struct_ptr=params)

        duration = pnlvm.helpers.load_extract_scalar_array_one(builder, duration_ptr)
        initial = pnlvm.helpers.load_extract_scalar_array_one(builder, initial_ptr)
        final = pnlvm.helpers.load_extract_scalar_array_one(builder, final_ptr)

        exp_f = ctx.get_builtin("exp", [ctx.float_ty])
        val = builder.load(ptri)
        val = builder.fmul(val, duration)
        val = builder.fadd(val, initial)
        val = builder.call(exp_f, [val])

        if "derivative" in tags:
            # f'(x) = s*r*e^(r*x + b)
            val = builder.fmul(val, final)
            val = builder.fmul(val, duration)
        else:
            # f(x) = s*e^(r*x + b) + o
            val = builder.fmul(val, final)

        builder.store(val, ptro)

    def _gen_pytorch_fct(self, device, context=None):
        initial = self._get_pytorch_fct_param_value(INITIAL, device, context)
        tolerance = self._get_pytorch_fct_param_value(TOLERANCE, device, context)
        final = self._get_pytorch_fct_param_value(FINAL, device, context)
        duration = self._get_pytorch_fct_param_value(DURATION, device, context)

        return lambda x : (initial - final) * torch.exp(x * torch.log(tolerance) / duration) + final



import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm


DIM_X=1500

# These are just basic tests to check that vector indexing and operations
# work correctly when compiled. The values don't matter much.
# Might as well make them representable in fp32 for single precision testing.
u = np.random.rand(DIM_X).astype(np.float32).astype(np.float64)
v = np.random.rand(DIM_X).astype(np.float32).astype(np.float64)
scalar = np.random.rand()

add_res = np.add(u, v)
sub_res = np.subtract(u, v)
mul_res = np.multiply(u, v)
smul_res = np.multiply(u, scalar)

@pytest.mark.benchmark(group="Hadamard")
@pytest.mark.parametrize("op, v, builtin, result", [
                         (np.add, v, "__pnl_builtin_vec_add", add_res),
                         (np.subtract, v, "__pnl_builtin_vec_sub", sub_res),
                         (np.multiply, v, "__pnl_builtin_vec_hadamard", mul_res),
                         (np.multiply, scalar, "__pnl_builtin_vec_scalar_mult", smul_res),
                         ], ids=["ADD", "SUB", "MUL", "SMUL"])
def test_vector_op(benchmark, op, v, builtin, result, func_mode):

    def _numpy_args(bin_f):
        np_u = u.astype(bin_f.np_arg_dtypes[0])
        np_v = bin_f.np_arg_dtypes[1].type(v) if np.isscalar(v) else v.astype(bin_f.np_arg_dtypes[1])
        np_res = np.empty_like(np_u)

        return np_u, np_v, np_res

    if func_mode == 'Python':
        def ex():
            return op(u, v)

    elif func_mode == 'LLVM':
        bin_f = pnlvm.LLVMBinaryFunction.get(builtin, ctype_ptr_args=(0, 1, 3))
        lu, lv, lres = _numpy_args(bin_f)

        ct_u = lu.ctypes.data_as(bin_f.c_func.argtypes[0])
        ct_v = lv if np.isscalar(lv) else lv.ctypes.data_as(bin_f.c_func.argtypes[1])
        ct_res = lres.ctypes.data_as(bin_f.c_func.argtypes[3])

        def ex():
            bin_f(ct_u, ct_v, DIM_X, ct_res)
            return lres

    elif func_mode == 'PTX':
        bin_f = pnlvm.LLVMBinaryFunction.get(builtin)
        lu, lv, lres = _numpy_args(bin_f)

        cuda_u = pnlvm.jit_engine.pycuda.driver.In(lu)
        cuda_v = lv if np.isscalar(lv) else pnlvm.jit_engine.pycuda.driver.In(lv)
        cuda_res = pnlvm.jit_engine.pycuda.driver.Out(lres)

        def ex():
            bin_f.cuda_call(cuda_u, cuda_v, np.int32(DIM_X), cuda_res)
            return lres

    res = benchmark(ex)
    np.testing.assert_allclose(res, result)


@pytest.mark.benchmark(group="Sum")
def test_vector_sum(benchmark, func_mode):

    if func_mode == 'Python':
        def ex():
            return np.sum(u)

    elif func_mode == 'LLVM':
        bin_f = pnlvm.LLVMBinaryFunction.get("__pnl_builtin_vec_sum", ctype_ptr_args=(0,))

        np_u = u.astype(bin_f.np_arg_dtypes[0])
        np_res = bin_f.np_buffer_for_arg(2)

        ct_u = np_u.ctypes.data_as(bin_f.c_func.argtypes[0])

        def ex():
            bin_f(ct_u, DIM_X, np_res)
            return np_res

    elif func_mode == 'PTX':
        bin_f = pnlvm.LLVMBinaryFunction.get("__pnl_builtin_vec_sum")

        np_u = u.astype(bin_f.np_arg_dtypes[0])
        np_res = bin_f.np_buffer_for_arg(2)

        cuda_u = pnlvm.jit_engine.pycuda.driver.In(np_u)
        cuda_res = pnlvm.jit_engine.pycuda.driver.Out(np_res)

        def ex():
            bin_f.cuda_call(cuda_u, np.int32(DIM_X), cuda_res)
            return np_res

    res = benchmark(ex)
    np.testing.assert_allclose(res, np.sum(u))
